PROGRAM PASCALCOMPILER(INPUT,OUTPUT);


 (* ORIGINAL VERSION : MARK D. RUSTAD 1977,
    VERSION OF P2-COMPILER ADAPTED FOR SMALL MACHINES *)


 (* V1.0 COMPILER BASED UPON PORTABLE P2 COMPILER
     WRITTEN BY N.WIRTH ET AL ZURICH E.T.H.              *)
 (* TYPED IN FROM SOURCE LISTING HANS OTTEN,
     1978 - 2020,
     - IMPROVED READABILITY
     - CHANGED UPPERCASE TO LOWERCASE AND CHANGED IT BACK
     - RESERVED WORDS PASCAL UPPERCASE
     - REMOVED CONFLICTS WITH RESERVED WORDS LIKE STRING   *)

CONST

  DISPLIMIT  =    15 ;
  MAXLEVEL   =    12 ;

 (* SIZE OF STORE ACCESSIBLE BY CODE *)
  MAXADDR    = 32767 ;

 (* SIZE IN BYTE IN CORE OF VARIABLES *)
 (* 8 BYTES FOR SET,GIVES 64 POSSIBLE MEMBERS *)
  INTSIZE    =     2 ;(* INTEGER   *)
  CHARSIZE   =     1 ;(* CHARACTER *)
  BOOLSIZE   =     1 ;(* BOOLEAN   *)
  SETSIZE    =     8 ;(* SET       *)
  PTRSIZE    =     2 ;(* POINTER   *)
  STRGLGTH   =     8 ;(* STRINGS STORED IN BLOCKS OF 8 CHAR'S *)
  MAXINT     = 32767 ;(* 16-BIT'S TWO'S COMPLEMENT MACHINE *)
  SETMAX     =    58 ;(* SET HAS MAXIMAL  MEMBERS *)

  LCAFTERMARKSTACK =  6 ;(* STACK FRAME FOR INTERPRETER *)

  MAXCHCNT   =   120 ;(* MAXIMUM INPUT LINE-LENGTH *)

  MAXSTANDRD = 10 ;   (* NR OF STANDARD PROCEDURES *)

 (* ASCII CHARACTER CONSTANTS *)
  AA 	   = 65 ;
  AF 	   = 70 ;
  AZ 	   = 90 ;
  A0       = 48 ;
  A9       = 57 ;
  APLUS    = 43 ;
  AMINUS   = 45 ;
  ASTAR    = 42 ;
  ASLASH   = 47 ;
  ALPAREN  = 40 ;
  ARPAREN  = 41 ;
  ADOLLAR  = 36 ;
  AEQUAL   = 61 ;
  ASPACE   = 32 ;
  ACOMMA   = 44 ;
  APERIOD  = 46 ;
  ADQUOTE  = 34 ;
  ALBRACK  = 91 ;
  ARBRACK  = 93 ;
  ACOLON   = 58 ;
  AARROW   = 94 ;
  ALESS    = 60 ;
  AGREATER = 62 ;
  ASEMI    = 59 ;
  ATAB     = 9   ;(* TAB-CHARACTER  *)
  AFORM    = 12  ;(* FORM-FEED *)

TYPE
 (* BASIC SYMBOLS *)
  SYMBOLN  =   (IDENT,
               INTCONST,
               STRINGCONST,
               NOTSY,
               MULOP,
               ADDOP,
               RELOP,
               LPARENT,
               RPARENT,
               LBRACK,
               RBRACK,
               COMMA,
               SEMICOLON,
               PERIOD,
               ARROW,
               COLON,
               BECOMES,
               CONSTSY,
               TYPESY,
               VARSY,
               FUNCSY,
               PROGSY,
               PROCSY,
               SETSY,
               PACKEDSY,
               ARRAYSY,
               RECORDSY,
               EXTERNSY,
               FORWARDSY,
               BEGINSY,
               IFSY,
               CASESY,
               REPEATSY,
               WHILESY,
               FORSY,
               ENDSY,
               ELSESY,
               UNTILSY,
               OFSY,
               DOSY,
               TOSY,
               DOWNTOSY,
               THENSY,
               OTHERSY  ) ;

  OPERATORN =  (MUL,
	       RDIV,
               ANDOP,
               IDIV,
               IMOD,
               PLUS,
               MINUS,
               OROP,
               LTOP,
               LEOP,
               GEOP,
               GTOP,
               NEOP,
               EQOP,
               INOP,
               NOOP ) ;

   SETOFSYS  =  SET OF SYMBOLN ;

   INTSET    =  SET OF 0 .. SETMAX ;

   STRINGT   = PACKED ARRAY[1 .. STRGLGTH] OF 0..127 ;


  (* CONSTANTS *)
   CSP = ^CONSTANT ;
   CONSTANT = RECORD
                 CASE INTEGER OF
                   (* SET *)
                    0 :(PVAL   : INTSET) ;
                   (* STRING *)
                    1 :(SLGTH : 0 .. STRGLGTH ;
                         (* LINKED LIST POINTER *)
                         SPTR  : CSP ;
                         SVAL  : STRINGT) 
                  END ;
  (* VALUE OF CONSTANT,CAN BE SIMPLE INTEGER OR COMPLEX *)
   VALU  =  RECORD
              CASE INTEGER OF
                0 :(IVAL : INTEGER) ;
                1 :(VALP : CSP    )
               END ;

 (* DATA STRUCTURES *)
  LEVRANGE  = 0 .. MAXLEVEL ;
  ADDRRANGE = 0 .. MAXADDR  ;
  STRUCTFORM =(SCALAR, SUBRANGE, POINTER, POWER, ARRAYS, RECORDS ) ;
  DECLKIND  = (STANDARD,
                 DECLARED) ;
  STP = ^STRUCTURE  ;
  CTP = ^IDENTIFIER ;

  STRUCTURE = PACKED RECORD
                SIZE  : ADDRRANGE ;
                STYPE : STP ;
                CASE FORM : STRUCTFORM OF
                  SUBRANGE :(MIN,MAX  : VALU      ) ;
                  ARRAYS   :(INXTYPE  : STP ) ;
                  RECORDS  :(FSTFLD   : CTP  )
               END ;
  (* NAMES,IDENTIFIER-CLASSES  *)
   IDCLASS =(TYPES, KONST, VARS, FIELD, PROC, FUNC) ;

   SETOFIDS = SET OF IDCLASS ;

   IDKIND =( ACTUAL,FORMAL) ;

   ALPHA = PACKED ARRAY[1 .. 8] OF CHAR ;

   IDENTIFIER = PACKED RECORD
                  NAME   : ALPHA ;
                  LLINK,
                  RLINK  : CTP  ;  (* TO BUILD BINARY TREE *)
                  IDTYPE : STP ;
                  NEXT   : CTP  ;  (* TO BUILD LINKED LIST *)
                  CASE KLASS : IDCLASS OF
				                 KONST :(VALUES : VALU)   ;
                                 VARS  :(VKIND : IDKIND    ;
                                 VLEV  : LEVRANGE  ;
                                 VADDR : ADDRRANGE) ;
                                 PROC,FUNC :
                                 (CASE PFDECKIND : DECLKIND OF
                                    STANDARD :(KEY : 1 .. 8) ;
                                    DECLARED :(PFLEV   : LEVRANGE ;
                                    PFNAME   : INTEGER  ;
                                    FORWDECL,
                                    EXTERNL  : BOOLEAN ) )
                 END ;
   DISPRANGE = 0 .. DISPLIMIT ;
  (* EXPRESSIONS *)
   ATTRKIND =(CST, VARBL, EXPR) ;

   VACCESS = (DRCT, INDRCT) ;
  (* RESULT OF EXPRESSION EVALUATION *)
   ATTR    =  RECORD
                TYPTR : STP ;
                CASE KIND : ATTRKIND OF
                  CST   :(CVAL : VALU) ;
                  VARBL :(ACCESS : VACCESS  ;
                            LEVEL  : LEVRANGE ;
                            DPLMT  : ADDRRANGE)
               END ;

   MARKPT = ^CHAR ;

 VAR
  (* VARIABLES RETURNED BY SOURCE PROGRAM SCANNER INSYMBOL *)
   SY   : SYMBOLN    ;         (* LAST SYMBOL FOUND *)
   OP   : OPERATORN  ;         (* CLASSIFICATION OF LAST SYMBOL *)
   VAL  : VALU       ;         (* VALUE OF LAST SYMBOL *)
   LGTH : INTEGER    ;         (* LENGTH OF LAST STRING CONSTANT *)
   ID   : ALPHA      ;         (* LAST IDENTIFIER,POSSIBLY TRUNCATED *)
   KK   : 0  .. 8    ;         (* NR'S OF CHARS IN LAST IDENTIFIER *)
   CH   : CHAR       ;         (* LAST CHARACTER READ *)
   ICH  : 0 .. 127   ;         (* ASCII VALUE OF LAST CHARACTER *)

  (*  LINE VARIABLES *)
   LINELENGTH,
   CHCNT        : 0 .. MAXCHCNT ;
   LINE         : PACKED ARRAY[1 .. MAXCHCNT] OF CHAR ;(* LAST SOURCE LINE *)
   LINECOUNT    : INTEGER ;

  (* COUNTERS *)
   LC,
   IC   : ADDRRANGE   ;    (* DATA LOCATION AND INSTRUCTION COUNTERS *)
   ICN,
   SUMCHECK : INTEGER ;    (* CHECKSUM *)
   CODEBUF  : ARRAY[0 .. 29] OF 0 .. 255 ;
   NPROC    : INTEGER ;    (* NEXT PROCEDURE NUMBER *)
   I,
   MXINT10  : INTEGER ;

  (* SWITCHES *)
   DP,                    (* DECLARATION PART *)
   PRTERR  : BOOLEAN ;     (* TO ALLOW FORWARD REFERENCE IN POINTER
                               TYPE DECLARATION BY SUPPRESSING ERROR
                               MESSAGE    *)
   PFCTTEST : BOOLEAN ;    (* DETECT FUNCTION CALLS IN
                               PARAMETER LISTS *)
   ERRFLAG  : BOOLEAN ;    (* COMPILER ERROR FLAG *)

  (* POINTERS *)
   INTPTR,
   CHARPTR,
   BOOLPTR,
   NILPTR   : STP  ;    (* POINTERS TO STANDARD IDENTIFIERS *)

   UTYPPTR,
   UCSPTR,
   UVARPTR,
   UFLDPTR,
   UPRCPTR,
   UFCTPTR  : CTP ;      (* POINTERS FOR UNDECLARED IDENTIFIERS *)
   FWPTR    : CTP ;      (* HEAD OF CHAIN OF FORW DECL TYPE IDS *)
   PROGPTR  : CTP ;      (* POINTER TO PROGRAM *)

  (* BOOKKEEPING OF DECLARATION LEVELS *)
   LEVEL   : LEVRANGE  ;   (* CURRENT STATIC LEVEL *)
   DISX,                  (* LEVEL OF LAST ID SEARCHED BY SEARCHID*)
   TOP     : DISPRANGE ;   (* TOP OF DISPLAY *)
   SAVETOP : DISPRANGE ;   (* SAVE TOP OF DISPLAY *)

  (* SYMBOL-TABLE IS ORGANIZED AS BINARY TREE'S PER LEVEL.
      EACH ENTRY IN DISPLAY-ARRAY POINTS TO UNBALANCED BINARY TREE,
      ARRAY-INDEX IS LEVEL OF CURRENT PROCEDURE/FUNCTION       *)
   DISPLAY : ARRAY[DISPRANGE] OF CTP;

  (* ERROR MESSAGES *)
   ERRINX  : 0 .. 10 ;      (* NUMBER OF ERRORS
                               IN CURRENT SOURCE LINE *)
   ERRLIST : ARRAY[1 .. 10] OF
               PACKED RECORD
                 POS : 1 .. 132  ;(* POSITION OF ERROR FOUND *)
                 NMR : 1 .. 401  (* NUMBER OF ERROR FOUND   *)
                END ;
  (* EXPRESSION COMPILATION *)
   GATTR : ATTR  ;   (* DESCRIBES THE EXPRESSION BEING COMPILED *)
  (* STRUCTURED CONSTANTS *)
   CONSTBEGSYS,
   SIMPTYPEBEGSYS,
   TYPEBEGSYS,
   BLOCKBEGSYS,
   SELECTSYS,
   FACBEGSYS,
   STATBEGSYS,
   TYPEDELS       :    SETOFSYS ;

   RW  : ARRAY[0 .. 31]  OF ALPHA    ; (* RESERVED WORDS       *)
   FRW : ARRAY[0 .. 8 ]  OF 0 .. 32  ; (* NO RES WRDS          *)
   RSY : ARRAY[0 .. 31]  OF  SYMBOLN ; (* RESERVED SYMBOLS     *)
   SSY : ARRAY[0 .. 127] OF SYMBOLN  ; (* NOT RESERVED SYMBOLS *)
   ROP : ARRAY[0 .. 31]  OF OPERATORN; (* RESERVED             *)
   SOP : ARRAY[0 .. 127] OF OPERATORN;
   NA  : ARRAY[0 .. 12]  OF ALPHA    ; (* STANDARD PROC/FUNC'S *)

(* EXTERNAL PROCEDURES,ONLY MEANINGFUL FOR PASCAL-M COMPILER   *)
PROCEDURE TTYIN  ; EXTERN = $00C0 ;(* SWITCH READ  TO TTY      *)
PROCEDURE SETHSR ; EXTERN = $00C8 ;(* SWITCH READ TO PAPTERTAPE*)
PROCEDURE TTYOUT ; EXTERN = $00D0 ;(* SWITCH WRITE   TO TTY    *)
PROCEDURE SETHSP ; EXTERN = $00D8 ;(* SWITCH WRITE TO PAPERTAPE*)

(* EXTERNAL PROCEDURES PASCAL-M  *)

(* DUMMY PASCAL-M PROCEDURES TO AID COMPILATION WITH OTHER PASCAL COMPILER 


PROCEDURE TTYIN  ;

   BEGIN
   END ;

PROCEDURE SETHSR ;

   BEGIN
   END ;

PROCEDURE TTYOUT ;

   BEGIN
   END ;

PROCEDURE SETHSP ;

   BEGIN
   END ;

PROCEDURE RELEASE(MARKP : MARKPT) ;

  BEGIN

  END ;
  END OF PASCAL-M PROCEDURES TO AID COMPILATION OTHER PASCAL COMPILER *)

  PROCEDURE BEGINLINE ;
 (* READS LINE FROM SOURCE INTO LINE *)

    BEGIN
      IF EOF
        THEN
          WRITELN('PREMATURE END OF SOURCE FILE') ;
      (* EMPTY LINE, SKIP *)
      WHILE EOLN DO
        BEGIN
          LINECOUNT := LINECOUNT + 1 ;
          READLN ;
        END;
      LINECOUNT := LINECOUNT + 1 ;
      LINELENGTH := 1 ; READ(LINE[LINELENGTH]) ;
      WHILE (NOT EOLN) AND (LINELENGTH < MAXCHCNT) DO
        BEGIN
          LINELENGTH := LINELENGTH + 1 ;
	  READ(LINE[LINELENGTH])
        END ;
      IF EOLN
        THEN
          READLN ;
      CHCNT := 0 ;
    END ;(* BEGINLINE *)

  PROCEDURE ENDLINE ;
  (*  REPORTS ERRORS IF ANY *)

   VAR
       LASTPOS,
       FREEPOS,
       CURRPOS,
       CURRNMR,
       F, K     : INTEGER ;

   BEGIN
     (* OUTPUT ERROR-MESSAGES *)
     IF ERRINX > 0
       THEN
         BEGIN
           TTYOUT ;
           WRITELN ;
           WRITE ( ' ', LINECOUNT:4) ;
           FOR K := 1 TO LINELENGTH DO
             WRITE(LINE[K]) ;
           WRITELN ;
           WRITE('****') ;
           LASTPOS := 0 ;
           FREEPOS := 1 ;
           FOR K :=1 TO ERRINX DO
             BEGIN
               CURRPOS := ERRLIST[K].POS ;
               CURRNMR := ERRLIST[K].NMR ;
               IF CURRPOS = LASTPOS
                 THEN
                   BEGIN
                     WRITE(',')  ;
                   END
                 ELSE
                   BEGIN
                     WHILE FREEPOS < CURRPOS DO
                       BEGIN
                         WRITE(' ') ;
                         FREEPOS := FREEPOS + 1 ;
                       END ;
                     WRITE('^') ;
                     LASTPOS := CURRPOS ;
                   END ;
                 IF CURRNMR < 10
                   THEN
                     F := 1
                   ELSE
                     IF CURRNMR < 100
                       THEN
                         F := 2
                       ELSE
                         F := 3 ;
                 WRITE(CURRNMR:F) ;
                 FREEPOS := FREEPOS + F + 1 ;
             END ;
           WRITELN ;
           ERRINX := 0 ;
         END ;
       SETHSP ;
     END ; (* ENDLINE *)

  PROCEDURE ERROR(FERRNR : INTEGER) ;
 (* ADDS ERROR TO ERRORS FOR CURRENT LINE *)

    BEGIN
      ERRFLAG := TRUE ;
      IF ERRINX >= 9
        THEN
          BEGIN
            ERRLIST[10].NMR := 255 ;
            ERRINX := 10
          END
        ELSE
          BEGIN
            ERRINX := ERRINX + 1 ;
            ERRLIST[ERRINX].NMR := FERRNR
          END ;
      ERRLIST[ERRINX].POS := CHCNT
    END ;(* ERROR *)

    FUNCTION STRINGSIZE(LVP : CSP) : INTEGER ;
   (* CALCULATES LENGTH OF STRING,BUILD UP OF LINKED LIST
       OF CTP'S CELLS,LVP POINTS AT LIST-HEAD  *)

      VAR
        N : INTEGER ;(* LENGTH COUNTER *)

      BEGIN
        N := 0 ;
       (* WALK THROUGH LINKED LIST UNTIL END *)
        REPEAT
          N := N + LVP^.SLGTH ;
          LVP := LVP^.SPTR      (* NEXT CELL *)
        UNTIL LVP = NIL ;
        STRINGSIZE := N         (* RETURN RESULT *)
    END ;(* STRINGSIZE *)

  PROCEDURE INSYMBOL ;
 (* READ NEXT BASIC SYMBOL OF SOURCE PROGRAM AND RETURN ITS
     DESCRIPTION IN THE GLOBAL VARIABLES
     SY   = SYMBOL FOUND
     OP   = OPERATOR IF ANY
     ID   = NAME IF IDENTIFIER
     VAL  = VALUE IF CONSTANT
     LGTH = NUMBER OF CHARACTERS IF STRING

     SYMBOLS PARSED :

     - IDENTIFIER,RESERVED OR DECLARED
     - INTEGER-CONSTANT
     - STRING-CONSTANT
     - COLON OR BECOMES
     - PERIOD OR DOUBLE PERIOD(=COLON)
     - OPERATOR-SYMBOLS LESS,LESS OR EQUAL,GREATER,GREATER OR EQUAL,
                      NOT EQUAL
     - HEX-CONSTANT,RETURNED AS INTEGER-CONSTANT
     - SINGLE CHARACTER SYMBOLS AS FOUND IN TABLE SSY

     - COMMENTS,NEW LINES  SKIPPED,ILLEGAL SYMBOLS AND
       CHARACTERS FLAGGED                                  *)

      VAR
          I,K : INTEGER ;
          CHSTRING : STRINGT ;
          LVP,LVP1 : CSP;
          TEST,COMNT : BOOLEAN ;
          QUOTECHAR : CHAR ;

      PROCEDURE NEXTCH ;
      (* RETURNS NEXT CHARACTER FROM INPUT IN
         CH  = CHARACTER,
         ICH = ORDINAL VALUE OF CHARACTER *)

        BEGIN
         (* NEW LINE FROM INPUT NECESSARY ? *)
          IF CHCNT > LINELENGTH
            THEN
              BEGIN
                ENDLINE ;
                BEGINLINE
              END ;
         CHCNT := CHCNT + 1 ;
         IF CHCNT > LINELENGTH
           THEN
             CH := ' '    (* RETURN BLANK IF END OF LINE REACHED *)
           ELSE
             CH := LINE[CHCNT] ;
          (* GET ORDINAL VALUE *)
          ICH := ORD(CH) ;
        END ; (* NEXTCH *)

  BEGIN (* INSYMBOL *)
     REPEAT(* UNTIL SYMBOL FOUND,SKIP COMMENTS *)
     COMNT := TRUE ;
    (* SKIP BLANKS AND TABS *)
     WHILE ( (CH = ' ') OR (CH = CHR(9)) ) DO
       NEXTCH ;
     IF(CH >= 'A') AND(CH <= 'Z')
       THEN
        (* IDENTIFIER FOUND,STARTS WITH UPPERCASE :
                           --------
          IDENT -->-------! LETTER !----------->--
                     ^     --------      V
                     !     --------      !
                      --<-! LETTER !-----
                     !     --------      !
                     !     --------      V
                      --<-! DIGIT  !-----
                          --------             *)
         BEGIN
           KK := 0 ;
           ID := '        ' ;
           REPEAT
            (* ONLY 8 CHARACTERS SIGNIFICANT IN NAME,SKIP REST *)
             IF KK < 8
               THEN
                 BEGIN
                   KK := KK + 1 ;
                   ID[KK]  := CH
                 END ;
             NEXTCH
           UNTIL NOT(((CH >= '0') AND (CH <= '9')) OR
                     ((CH >= 'A') AND (CH <= 'Z')) ) ;
           SY := IDENT ;
           OP := NOOP  ;
          (* SEARCH FOR RESERVED NAME,FAST SEARCH BY USING
              LENGTH AS INDEX WITH TABLE KK                  *)
           I := FRW[KK-1] ;
           WHILE I <= FRW[KK] -1 DO
             BEGIN
               IF RW[I] = ID
                 THEN
                   BEGIN
                     SY := RSY[I] ;
                     OP := ROP[I]
                   END ;
               I := I + 1
             END
         END
       ELSE IF(CH >= '0') AND(CH <= '9')
         THEN
            (* INTEGER CONSTANT,ONLY DIGITS EXPECTED :

                               --------
      INTEGER-CONSTANT ->-----! DIGIT  !----------->--
                         ^     --------      V
                         !     --------      !
                          --<-! DIGIT  !-----
                               --------             *)

           BEGIN
             OP := NOOP ;
             TEST := FALSE ;
             K := 0 ;
             REPEAT
               IF K <= MXINT10
                 THEN
                   K := K * 10 + ICH - ORD('0')
                 ELSE
                   TEST := TRUE ;
               NEXTCH
             UNTIL(CH < '0') OR(CH > '9') ;
             IF TEST
               THEN
                 ERROR(203) ;(* NUMBER TOO LARGE *)
             VAL.IVAL := K ;
             SY := INTCONST
           END
        ELSE IF(CH = '''') OR(CH = '"')
         THEN
          (* STRING,STARTS WITH AND ENDS WITH(D)QUOTE *)
          (* PUT STRING FROM INPUT IN LINKED LIST OF
              CONSTANTS,EACH CONSTANT CONTAINING MAX 8 CHAR'S *)
           BEGIN
             LGTH := 0 ;
             SY := STRINGCONST ;
             OP := NOOP ;
             LVP := NIL ;
             QUOTECHAR := CH ;(* START AND END WITH SAME QUOTE *)
             REPEAT
               REPEAT
                 IF(CHCNT > LINELENGTH)
                   THEN
                     ERROR(202) ; (* STRING MUST BE ON ONE LINE *)
                 NEXTCH ;
                 LGTH := LGTH + 1 ;
                 IF LGTH >STRGLGTH
                   THEN
                     BEGIN
                       NEW(LVP1) ;
                       IF LVP <> NIL
                         THEN
                           LVP^.SPTR := LVP1 (* LINK TO LIST *)
                         ELSE
                           VAL.VALP := LVP1 ;(* NEW LIST *)
                       LVP := LVP1 ;
                       LVP^.SVAL := CHSTRING ;
                       LVP^.SLGTH := STRGLGTH ;
                       LVP^.SPTR := NIL ;
                       LGTH := 1
                     END ;
                 CHSTRING[LGTH] := ICH
               UNTIL CH = QUOTECHAR ;
               NEXTCH
             UNTIL CH <> QUOTECHAR ;
             LGTH := LGTH - 1 ;     (* NR OF CHAR'S IN STRING *)
             IF(LGTH = 1) AND(LVP = NIL)
               THEN
                 VAL.IVAL := CHSTRING[1](* ONLY ONE CHARACTER *)
               ELSE
                (* CREATE LIST-HEAD POINTER DESCRIBING STRING *)
                 IF LGTH <> 0
                   THEN
                     BEGIN
                       NEW(LVP1) ;
                       IF LVP <> NIL
                         THEN
                           LVP^.SPTR := LVP1
                         ELSE
                           VAL.VALP := LVP1 ;
                       LVP1^.SVAL := CHSTRING ;
                       LVP1^.SLGTH := LGTH  ;
                       LVP1^.SPTR := NIL ;
                       LGTH := STRINGSIZE(VAL.VALP) ;
                     END
           END
        ELSE IF CH = ':'
         THEN
           BEGIN
             OP := NOOP ;
             NEXTCH ;
             IF CH = '='
               THEN
                 BEGIN
                   SY := BECOMES ;
                   NEXTCH
                 END
               ELSE
                 SY := COLON
             END

       ELSE IF CH = '.'
         THEN
           BEGIN
             OP := NOOP ;
             NEXTCH ;
             IF CH = '.'
               THEN
                 BEGIN
                   SY := COLON ;
                   NEXTCH
                 END
               ELSE
                 SY := PERIOD
             END
       ELSE IF CH = '<'
         THEN
           BEGIN
             NEXTCH ;
             SY := RELOP ;
             IF CH = '='
               THEN
                 BEGIN
                   OP := LEOP ;
                   NEXTCH
                 END
               ELSE
                 IF CH = '>'
                   THEN
                     BEGIN
                       OP := NEOP ;
                       NEXTCH
                     END
                   ELSE
                     OP := LTOP
            END
       ELSE IF CH = '>'
         THEN
           BEGIN
             NEXTCH ;
             SY := RELOP ;
             IF CH = '='
               THEN
                 BEGIN
                   OP := GEOP ;
                   NEXTCH
                 END
               ELSE
                 OP := GTOP
           END
       ELSE IF CH = '('
         THEN
           BEGIN
             NEXTCH ;
             IF CH = '*'(* SKIP COMMENT *)
               THEN
                 BEGIN
                   NEXTCH ;
                   REPEAT
                     WHILE CH <> '*' DO
                       NEXTCH ;
                     NEXTCH ;
                   UNTIL CH = ')' ;
                   NEXTCH ;
                   COMNT := FALSE
                 END
               ELSE
                 BEGIN
                   SY := LPARENT ;
                   OP := NOOP
                 END
           END
       ELSE IF CH = '{'(* COMMENT *)
         THEN
           BEGIN
            (* SKIP UNTIL RIGHT ACCULADE FOUND *)
             WHILE CH <> '}' DO
               NEXTCH ;
             NEXTCH ;
             COMNT := FALSE
           END
       ELSE IF CH = '$'(* HEX CONSTANT,0 .. 9,A .. F *)
         THEN
           BEGIN
             OP := NOOP ;
             K := 0 ;
             REPEAT
               NEXTCH ;
               IF(CH >= '0') AND(CH <= '9')
                 THEN
                   I := ICH - ORD('0')
               ELSE IF(CH >= 'A') AND(CH <= 'F')
                 THEN
                   I := ICH - ORD('A') + 10
               ELSE IF(CH >= 'A') AND(CH <= 'F')
                 THEN
                   I := ICH - ORD('A') + 10
               ELSE
                       I := -1 ;
               IF I >= 0
                 THEN
                   K := 16*K + I ;
             UNTIL I < 0 ;
             VAL.IVAL := K ;
             SY := INTCONST
           END
       ELSE (* OTHER SINGLE CHARACTER SYMBOLS *)
         BEGIN
           SY := SSY[ICH] ;
           OP := SOP[ICH] ;
           NEXTCH ;
           IF SY = OTHERSY
             THEN
               ERROR(399) (* ILLEGAL CHARACTER FOUND *)
          END ;

    UNTIL COMNT ;

  END ;(* INSYMBOL *)

  PROCEDURE ENTERID(FCP : CTP) ;
 (* ENTER ID POINTED AT BY FCP INTO THE NAME-TABLE,
     WHICH ON EACH DECLARATION LEVEL IS ORGANIZED AS
     AN UNBALANCED BINARY TREE                       *)

  VAR
    NAM : ALPHA ;
    LCP,LCP1 : CTP;
    LLEFT : BOOLEAN ;

  BEGIN
    NAM := FCP^.NAME ;
    LCP := DISPLAY[TOP] ;
    IF LCP = NIL
      THEN
       (* EMPTY TREE,ENTER AT ROOT *)
        DISPLAY[TOP] := FCP
      ELSE
        BEGIN
         (* FIND PLACE IN TREE WHERE TO PUT IDENTIFIER *)
          REPEAT
            LCP1 := LCP ;
           (* DECIDE IF LEFT OR RIGHT TO GO,IF
               NAME CONFLICT FOLLOW RIGHT LINK *)
            IF LCP^.NAME = NAM
              THEN
                BEGIN
                 (* ALREADY PRESENT *)
                  ERROR(101) ;
                  LCP := LCP^.RLINK ;
                  LLEFT := FALSE
                END
            ELSE IF LCP^.NAME < NAM
              THEN
                BEGIN
                 (* GO RIGHT *)
                  LCP := LCP^.RLINK ;
                  LLEFT := FALSE
                END
            ELSE
              BEGIN
               (* GO LEFT *)
                LCP := LCP^.LLINK ;
                LLEFT := TRUE
              END
          UNTIL LCP = NIL ;

         (* PLACE FOUND,LINK IDENTIFIER IN TREE *)
          IF LLEFT
            THEN
              LCP1^.LLINK := FCP
            ELSE
              LCP1^.RLINK := FCP
        END ;
   (* IDENTIFIER IS NEW LEAVE,NO LEFT OR RIGHT POINTERS YET *)
    FCP^.LLINK := NIL ;
    FCP^.RLINK := NIL
  END ; (* ENTERID *)

  PROCEDURE SEARCHSECTION(FCP : CTP; VAR FCP1 : CTP) ;
 (* TO FIND RECORD FIELDS AND FORWARD DECLARED
     PROCEDURE ID'S
     CALLED BY
        - PROCEDURE PROCEDUREDECLARATION
        - PROCEDURE SELECTOR
     SEARCH STARTS IN TREE AT IDENTIFIER POINTED BY FCP,
     SEARCHED FOR IDENTIFIER WITH SAME NAME AS ID        *)

  BEGIN
   (* INDICATE NOT FOUND YET *)
    FCP1 := NIL ;
    WHILE(FCP <> NIL) AND(FCP1 = NIL) DO
      IF FCP^.NAME <> ID
       THEN
        (* NOT FOUND,DESCEND DOWN RIGHT OR LEFT *)
         IF FCP^.NAME < ID
           THEN
             FCP := FCP^.RLINK
           ELSE
             FCP := FCP^.LLINK
      ELSE
      (* FOUND,STOP SEARCH AND RETURN POINTER *)
         FCP1 := FCP
  END ;(* SEARCHSECTION *)

  PROCEDURE SEARCHID(FIDCLS : SETOFIDS ; VAR FCP : CTP) ;
 (* SEARCH IDENTIFIER IN SYMBOL-TABLE,START AT DISPLAY-LEVEL
     TOP AND GO DOWN TO BOTTOM DISPLAY. SEARCH BASED ON NAME ID.
     IF FOUND THEN TYPE IS CHECKED WITH TYPES IN FIDCLS.
     IF NOT FOUND THEN RETURN UNKNOWN TYPE-DESCRIPTION *)

  VAR
    LCP : CTP;

  BEGIN
 (* SEARCH STARTS AT TOP AND IDENTIFIER IS NOT FOUND YET *)
  DISX := TOP + 1 ;
  FCP := NIL ;
  WHILE(DISX > 0) AND(FCP = NIL) DO
    BEGIN
      DISX := DISX - 1 ;
     (* NEW ROOT *)
      LCP := DISPLAY[DISX] ;
     (* SEARCH TREE *)
      WHILE(LCP <> NIL) AND(FCP = NIL) DO
       (* DECIDE IF FOUND OR GO DOWN RIGHT OR LEFT *)
        IF LCP^.NAME = ID
          THEN
           (* CHECK IF TYPE IS IN FIDCLS *)
            IF LCP^.KLASS IN FIDCLS
              THEN
                FCP := LCP
              ELSE
                BEGIN
                 (* NOT APPROPIATE CLASS *)
                  IF PRTERR
                    THEN
                      ERROR(103) ;
                 (* GO RIGHT,MAYBE DOUBLE DECLARATION *)
                  LCP := LCP^.RLINK
                END
        ELSE IF LCP^.NAME < ID
          THEN
           (* GO RIGHT *)
            LCP := LCP^.RLINK
        ELSE
         (* GO LEFT *)
          LCP := LCP^.LLINK
    END ;
   (* SEARCH NOT SUCCESSFUL,SUPPRESS ERROR MESSAGE
       IN CASE OF FORWARD REFERENCED TYPE ID IN POINTER TYPE
       DEFINITION --> PROCEDURE SIMPLETYPE                  *)
    IF PRTERR AND(FCP = NIL)
      THEN
        BEGIN
          ERROR(104) ;
         (* TO AVOID RETURNING NIL,REFERENCE AN ENTRY
             FOR UNDECLARED ID OF APPROPIATE CLASS
             ---> PROCEDURE ENTERUNDECLARED                   *)
          IF TYPES IN FIDCLS
            THEN
              FCP := UTYPPTR
          ELSE IF VARS IN FIDCLS
            THEN
              FCP := UVARPTR
          ELSE IF FIELD IN FIDCLS
            THEN
              FCP := UFLDPTR
          ELSE IF KONST IN FIDCLS
            THEN
              FCP := UCSPTR
          ELSE IF PROC IN FIDCLS
            THEN
              FCP := UPRCPTR
          ELSE
            FCP := UFCTPTR ;
        END
  END ;(* SEARCHID *)

  PROCEDURE GETBOUNDS(FSP : STP ; VAR FMIN,FMAX : INTEGER) ;
 (* GET INTERNAL BOUNDS OF SUBRANGE OR SCALER TYPE,
     ASSUME FSP <> NIL AND FSP^.FORM <= SUBRANGE AND FSP <> INTPTR,
     RETURNS BOUNDS IN FMIN AND FMAX *)

  BEGIN
    IF FSP^.FORM = SUBRANGE
      THEN
        BEGIN
          FMIN := FSP^.MIN.IVAL ;
          FMAX := FSP^.MAX.IVAL
        END
      ELSE  (* SCALAR *)
        BEGIN
          FMIN := 0 ;
          IF FSP = CHARPTR
            THEN
             (* CHARACTERS *)
              FMAX := 127
            ELSE
             (* OTHER SCALARS *)
              FMAX := 0
        END
  END ;(* GETBOUNDS *)

  PROCEDURE HEXOUT(WORD : INTEGER) ;
 (* WRITE WORD AS FOUR HEX CHARACTERS *)

  VAR
    K : INTEGER ;

  BEGIN
    SUMCHECK := SUMCHECK + WORD ;
    WHILE WORD < 0 DO
      WORD := WORD + 256 ;
    K := WORD DIV 16 ;
    WORD := WORD MOD 16 ;
    IF K < 10
      THEN
        WRITE(CHR(ORD('0') + K))
      ELSE
        WRITE(CHR(ORD('A') + K - 10));
    IF WORD < 10
      THEN
        WRITE(CHR(ORD('0') + WORD))
      ELSE
        WRITE(CHR(ORD('A') + WORD  - 10)) ;
  END ;(* HEXOUT *)

  PROCEDURE WRITEOUT ;
 (* WRITE P1-RECORD WITH CODE BUFFER AND CHECKSUM,
    CODEBUFFER EMPTY AND INSTRUCTION-COUNTER IS UPDATED *)

  VAR
    I : INTEGER ;

  BEGIN
    IF ICN <> 0
      THEN
        BEGIN
          WRITE('P1') ;
          SUMCHECK := 0 ;
          HEXOUT(ICN + 1) ;
          FOR I := 0 TO ICN - 1 DO
            HEXOUT(CODEBUF[I]) ;
          HEXOUT((16383 - SUMCHECK) MOD 256) ;
          WRITELN ;
         END ;
      IC := IC + ICN ;
      ICN := 0
  END ;(* WRITEOUT *)

  PROCEDURE BLOCK(FSYS : SETOFSYS ; FSY : SYMBOLN ; FPROCP : CTP);
 (* COMPILES BLOCK *)

  TYPE
    OPRANGE = 0..255 ;

  VAR
    LSY : SYMBOLN ;
    TEST : BOOLEAN ;

    PROCEDURE BYTEGEN (BYTE : INTEGER) ;
   (* ADD BYTE TO CODEBUFFER,EMPTY CODEBUFFER IF NECESSARY *)

    BEGIN
      IF ICN = 30
        THEN
          WRITEOUT ;
      CODEBUF[ICN] := BYTE ;
      ICN := ICN + 1
    END ;(* BYTEGEN *)

    PROCEDURE ADDRGEN(ADDR : INTEGER) ;
   (* PUT ADDR AS TWO BYTES IN CODEBUFFER *)

    BEGIN
      BYTEGEN(ADDR DIV 256) ;
      BYTEGEN(ADDR MOD 256)
    END ;(* ADDRGEN *)

    PROCEDURE GENUJPENT(FOP : OPRANGE ; FP2 : INTEGER) ;
   (* GENERATE FOP INSTRUCTION,FOLLOWED BY FP2 AS WORD *)

    BEGIN
      BYTEGEN(FOP) ;
      ADDRGEN(FP2)
    END ;(* GENUJPENT *)

    PROCEDURE PLANTADDR(LOC,VAL : INTEGER) ;
   (* PLANT VALUE AT LOCATION IN CODE,
       GENERATE P2-RECORD IF LOC NOT IN CODEBUFFER *)

    BEGIN
      IF VAL <> 0
        THEN
          IF(LOC >= IC) AND(LOC - IC < ICN)
            THEN
              BEGIN
               (* PLACE IN CODEBUFFER *)
               (* CORRECT FOR OFFSET IN CODEBUFFER *)
                LOC := LOC - IC ;
                CODEBUF[LOC]     := VAL DIV 256 ;
                CODEBUF[LOC + 1] := VAL MOD 256
              END
            ELSE
              BEGIN
               (* GENERATE P2-RECORD WITH LOCATION AND VALUE *)
                WRITEOUT ;
                WRITE('P2') ;
                SUMCHECK := 0 ;
               (* FIRST LOCATION *)
                HEXOUT(LOC DIV 256) ;
                HEXOUT(LOC MOD 256) ;
               (* NEXT VALUE TO BE PLANTED *)
                HEXOUT(VAL DIV 256) ;
                HEXOUT(VAL MOD 256) ;
                HEXOUT((16383 - SUMCHECK) MOD 256) ;
                WRITELN
              END ;
    END ;(* PLANTADDR *)

    PROCEDURE SKIP(FSYS : SETOFSYS) ;
   (* SKIP INPUT CHSTRING UNTIL RELEVANT SYMBOL FOUND *)

    BEGIN
      WHILE(NOT(SY IN FSYS)) DO
        INSYMBOL
    END ;(* SKIP *)

    PROCEDURE TEST1(FSYS : SETOFSYS ; N : INTEGER) ;
   (* TEST IF CURRENT SYMBOL SY IN SET FSYS,IF NOT GENERATE
       ERROR WITH NUMBER N ERRORNR AND SKIP UNTIL SY IN FSYS *)

    BEGIN
      IF NOT(SY IN FSYS)
        THEN
          BEGIN
            ERROR(N) ;
            SKIP(FSYS)
          END
    END ;(* TEST1 *)

    PROCEDURE TEST2(FSYS : SETOFSYS; N: INTEGER; S2 : SETOFSYS);
   (* TEST IF CURRENT SYMBOL SY IN SET FSYS,IF NOT GENERATE
       ERROR WITH NUMBER N ERRORNR AND SKIP UNTIL SY IN FSYS + S2 *)

    BEGIN
      IF NOT(SY IN FSYS)
        THEN
          BEGIN
            ERROR(N) ;
            SKIP(FSYS + S2)
          END
    END ;(* TEST2 *)

    PROCEDURE INTEST(SYM : SYMBOLN ; N : INTEGER) ;
   (* TEST IF CURRENT SYMBOL SY IS SYM,IF NOT GENERATE
       ERROR WITH NUMBER N ERRORNR,GET NEXT SYMBOL FROM INPUT *)

    BEGIN
      IF SY = SYM
        THEN
          INSYMBOL
        ELSE
          ERROR(N)
    END ;(* INTEST *)

    PROCEDURE CONSTANT(FSYS : SETOFSYS    ;
                       VAR FSP : STP      ;
                       VAR FVALU : VALU) ;
   (* CHECKS IF CONSTANT,DECLARED OR STRING-CONSTANT OR INTEGER,
       RETURNS DESCRIPTION IN FSP,VALUE IN FVALU *)

    TYPE
      SIGNS =(NONE,POS,NEG) ;

    VAR
       LSP  : STP ;
       LCP  : CTP;
       SIGN : SIGNS ;

    BEGIN
     (* NOT A CONSTANT,VALUE ZERO *)
      LSP := NIL ;
      FVALU.IVAL := 0 ;
      TEST2(CONSTBEGSYS,50,FSYS) ;
      IF SY IN CONSTBEGSYS
        THEN
          BEGIN
            IF SY = STRINGCONST
              THEN
                BEGIN
                  IF LGTH = 1
                    THEN
                      LSP := CHARPTR
                    ELSE
                      BEGIN
                        NEW(LSP) ;
                        LSP^.STYPE := CHARPTR ;
                        LSP^.INXTYPE := NIL ;
                        LSP^.SIZE := STRINGSIZE(VAL.VALP) ;
                        LSP^.FORM := ARRAYS ;
                      END ;
                  FVALU := VAL ;
                  INSYMBOL
                END
              ELSE
               (* OTHER THAN STRING-CONSTANT,FIRST CHECK IF SIGN *)
                BEGIN
                  SIGN := NONE ;
                  IF(SY = ADDOP) AND(OP IN[PLUS,MINUS])
                    THEN
                      BEGIN
                        IF OP = PLUS
                          THEN
                            SIGN := POS
                          ELSE
                            SIGN := NEG ;
                        INSYMBOL
                      END ;
                 (* CONSTANT CAN BE DECLARED IDENTIFIER OR DECLARED CONSTANT *)
                  IF SY = IDENT
                    THEN
                      BEGIN
                        SEARCHID([KONST],LCP) ;
                        LSP := LCP^.IDTYPE ;
                        FVALU := LCP^.VALUES ;
                        IF SIGN <> NONE
                          THEN
                            IF LSP = INTPTR
                              THEN
                                BEGIN
                                  IF SIGN = NEG
                                    THEN
                                      FVALU.IVAL := - FVALU.IVAL ;
                                END
                              ELSE
                               (* SIGN NOT ALLOWED *)
                                ERROR(105) ;
                        INSYMBOL
                      END
                  ELSE IF SY = INTCONST
                    THEN
                      BEGIN
                        IF SIGN = NEG
                          THEN
                            VAL.IVAL := -VAL.IVAL ;
                        LSP := INTPTR ;
                        FVALU := VAL ;
                        INSYMBOL
                      END
                  ELSE
                    BEGIN
                     (* NUMBER EXPECTED *)
                      ERROR(106) ;
                      SKIP(FSYS)
                    END
                END ;

            TEST1(FSYS,6)
          END ;
      FSP := LSP
    END ;(* CONSTANT *)

    FUNCTION COMPTYPES(FSP1,FSP2 : STP) : BOOLEAN ;
   (* DECIDE WHETHER STRUCTURES POINTED AT BY FSP1 AND
       FSP2 ARE COMPATIBLE                              *)
    BEGIN
      IF FSP1 = FSP2
        THEN
          COMPTYPES := TRUE
        ELSE
          IF(FSP1 <> NIL) AND(FSP2 <> NIL)
            THEN
              BEGIN
                IF FSP1^.FORM = FSP2^.FORM
                  THEN
                    CASE FSP1^.FORM OF
                      SCALAR,RECORDS          :
                        COMPTYPES := FALSE ;
                      SUBRANGE,POINTER,POWER  :
                         COMPTYPES :=
                             COMPTYPES(FSP1^.STYPE,FSP2^.STYPE);
                      ARRAYS                    :
                         COMPTYPES := COMPTYPES(FSP1^.STYPE,FSP2^.STYPE)
                                      AND(FSP1^.SIZE = FSP2^.SIZE)
                      END
                 (* FSP1^.FORM <> FSP2^.FORM) *)
                  ELSE IF FSP1^.FORM = SUBRANGE
                    THEN
                      COMPTYPES := COMPTYPES(FSP1^.STYPE,FSP2)
                  ELSE IF FSP2^.FORM = SUBRANGE
                     THEN
                       COMPTYPES := COMPTYPES(FSP1,FSP2^.STYPE)
                  ELSE
                    COMPTYPES := FALSE
            END
          ELSE
              COMPTYPES := TRUE
    END ;(* COMPTYPES *)

    FUNCTION ISSTRING(FSP : STP) : BOOLEAN ;
   (* TESTS IF STRUCTURE POINTED BY WITH FSP IS STRING *)

    BEGIN
      ISSTRING := FALSE ;
      IF FSP <> NIL
        THEN
          IF FSP^.FORM = ARRAYS
            THEN
              IF COMPTYPES(FSP^.STYPE,CHARPTR)
                THEN
                  ISSTRING := TRUE
    END ;(* ISSTRING *)

    PROCEDURE TYP(FSYS : SETOFSYS ;
                    VAR FSP : STP ;
                    VAR FSIZE : ADDRRANGE ) ;

    VAR
      LSP,LSP1,LSP2 : STP ;
      OLDTOP : DISPRANGE ;
      LCP : CTP;
      LSIZE,DISPL : ADDRRANGE ;
      LMIN,LMAX : INTEGER ;

    PROCEDURE SIMPLETYPE(FSYS : SETOFSYS ;
                           VAR FSP : STP ;
                           VAR FSIZE : ADDRRANGE) ;

   (* PARSES AND RETURNS DESCRIPTION AND SIZE OF SIMPLE TYPE
       STRUCTURE PARSED :

                          -----------------
       --------->--------! TYPE IDENTIFIER !--------------->
             !            -----------------            !
             V                                         ^
             !       ---     ------------     ---      !
             !-->---(()---! IDENTIFIER !---())-----!
             !       ---  !  ------------     ---      !
             !            !              !             !
             !            !      ---     !             !
             !             -----(,)----              !
             !                   ---                   !
             !                                         !
             !     ----------    ----    ----------    !
              ----! CONSTANT !--(..)--! CONSTANT !---
                   ----------    ----    ----------           *)

    VAR
      LSP,LSP1 : STP ;
      LCP,LCP1 : CTP;
      TTOP : DISPRANGE ;
      LCNT : INTEGER ;         (* KEEPS TRACK OF ENUMERATED CONSTANTS *)
      LVALU : VALU ;

    BEGIN
      FSIZE := 1 ;
      TEST2(SIMPTYPEBEGSYS,1,FSYS) ;
      IF SY IN SIMPTYPEBEGSYS
        THEN
          BEGIN
            IF SY = LPARENT
              THEN
               (* ENUMERATION TYPE *)
                BEGIN
                  TTOP := TOP ;
                  TOP := SAVETOP ;
                  NEW(LSP) ;
                  LSP^.FORM := SCALAR ;
                  LCP1 := NIL ;
                  LCNT := 0 ;
                  REPEAT
                    INSYMBOL ;
                   (* ENUMERATION MUST BE LIST OF CONSTANT IDENTS *)
                    IF SY = IDENT
                      THEN
                        BEGIN
                          NEW(LCP) ;
                          LCP^.NAME := ID ;
                          LCP^.IDTYPE := LSP ;
                          LCP^.NEXT := LCP1 ;
                          LCP^.VALUES.IVAL := LCNT ;
                          LCP^.KLASS := KONST ;
                          ENTERID(LCP) ;
                          LCNT := LCNT + 1 ;
                          LCP1 := LCP ;
                          INSYMBOL ;
                        END
                      ELSE
                        ERROR(2) ;
                    TEST1(FSYS +[COMMA,RPARENT],6) ;
                  UNTIL SY <> COMMA ;
                  LSP^.FSTFLD := LCP1 ;
                  TOP := TTOP ;
                 (* CHECK IF ENUMERATION-TYPE FITS IN ONE BYTE *)
                  IF LCNT < 257
                    THEN
                      LSP^.SIZE := 1
                    ELSE
                      LSP^.SIZE := INTSIZE ;
                  INTEST(RPARENT,4)
                END
            ELSE IF SY = IDENT
              THEN
               (* TYPE-IDENTIFIER OR DECLARED CONSTANT *)
                BEGIN
                  SEARCHID([TYPES,KONST],LCP) ;
                  INSYMBOL ;
                  IF LCP^.KLASS = KONST
                    THEN
                     (* SUBRANGE STARTS WITH CONSTANT *)
                      BEGIN
                        NEW(LSP) ;
                        LSP^.STYPE := LCP^.IDTYPE ;
                        LSP^.FORM := SUBRANGE ;
                        IF ISSTRING(LSP^.STYPE)
                          THEN
                            BEGIN
                              ERROR(148) ;
                              LSP^.STYPE := NIL
                            END ;
                        LSP^.MIN := LCP^.VALUES ;
                        LSP^.SIZE  := LCP^.IDTYPE^.SIZE ;
                        INTEST(COLON,5) ;
                        CONSTANT(FSYS,LSP1,LVALU) ;
                        LSP^.MAX := LVALU ;
                        IF LSP^.STYPE <> LSP1
                          THEN
                            ERROR(107)
                      END
                    ELSE
                      LSP := LCP^.IDTYPE
                  END(* SY IS IDENT *)
              ELSE(* MUST BE SUBRANGE STARTING WITH A CONSTANT *)
               BEGIN
                 NEW(LSP) ;
                 LSP^.FORM := SUBRANGE ;
                 CONSTANT(FSYS +[COLON],LSP1,LVALU) ;
                 IF ISSTRING(LSP1)
                   THEN
                     BEGIN
                       ERROR(148) ;
                       LSP1 := NIL
                     END ;
                  LSP^.STYPE := LSP1 ;
                  LSP^.MIN := LVALU ;
                  LSP^.SIZE := INTSIZE ;
                  INTEST(COLON,5) ;
                  CONSTANT(FSYS,LSP1,LVALU) ;
                  LSP^.MAX := LVALU ;
                  IF LSP^.STYPE <> LSP1
                    THEN
                      ERROR(107) ;
                 END ;
           IF LSP <> NIL
             THEN
               IF LSP^.FORM = SUBRANGE
                 THEN
                   IF LSP^.MIN.IVAL > LSP^.MAX.IVAL
                     THEN
                      (* STARTVALUE OF SUBRANGE MUST BE SMALLER
                          THAN ENDVALUE *)
                       ERROR(102) ;
            FSP := LSP ;
            TEST1(FSYS,6)
          END
        ELSE
          FSP := NIL ;
      IF LSP <> NIL
        THEN
          FSIZE := LSP^.SIZE ;
    END ;(* SIMPLETYPE *)


    PROCEDURE FIELDLIST(FSYS : SETOFSYS) ;
    (* PARSES FIELD-LIST,
       RETURNS DISPL = TOTAL SPACE NEEDED TO STORE FIELD-LIST IN RECORD
       USES NXT^ AS LINKED LIST,SEE JENSEN & WIRTH PAGE 116
       FOR SYNTAX DIAGRAM *)

    VAR
      LCP,LCP1,NXT,NXT1 : CTP;
      LSP,LSP3 : STP ;
      MINSIZE,MAXSIZE,LSIZE : ADDRRANGE ;
      LVALU : VALU ;

    BEGIN
      NXT1 := NIL ;
      LSP := NIL ;
      TEST2([IDENT,CASESY],19,FSYS ) ;
      WHILE SY = IDENT DO
        BEGIN
          NXT := NXT1 ;
          REPEAT
            IF SY = IDENT
              THEN
                BEGIN
                  NEW(LCP) ;
                  LCP^.NAME := ID ;
                  LCP^.IDTYPE := NIL ;
                 (* LINK TO LIST *)
                  LCP^.NEXT := NXT ;
                  LCP^.KLASS := FIELD ;
                  NXT := LCP ;
                  ENTERID(LCP) ;
                  INSYMBOL
                END
              ELSE
                ERROR(2) ;
            TEST2([COMMA,COLON],6,FSYS +[SEMICOLON,CASESY]) ;
            TEST := SY <> COMMA ;
            IF NOT TEST
              THEN
                INSYMBOL
          UNTIL TEST ;
          INTEST(COLON,5) ;
          TYP(FSYS +[CASESY,SEMICOLON],LSP,LSIZE) ;
          WHILE NXT <> NXT1 DO
            BEGIN
              NXT^.IDTYPE := LSP ;
              NXT^.VADDR := DISPL ;
              NXT := NXT^.NEXT ;
              DISPL := DISPL + LSIZE
            END ;
          NXT1 := LCP ;
          IF SY = SEMICOLON
            THEN
              BEGIN
                INSYMBOL ;
                TEST2([IDENT,CASESY,SEMICOLON],19,FSYS)
              END
        END ;(* WHILE *)
      NXT := NIL ;
      WHILE NXT1 <> NIL DO
        BEGIN
          LCP := NXT1^.NEXT ;
          NXT1^.NEXT := NXT ;
          NXT := NXT1 ;
          NXT1 := LCP
        END ;
     (* FIELD-LIST MAY CONTAIN A VARIANT PART *)
      IF SY = CASESY
        THEN
          BEGIN
            INSYMBOL ;
            IF SY = IDENT
              THEN
                BEGIN
                 (* SUPPRESS ERROR MESSAGES IF FORWARD REF POINTER *)
                  PRTERR := FALSE ;
                  SEARCHID([TYPES],LCP1) ;
                  PRTERR := TRUE ;
                  IF LCP1 = NIL
                    THEN
                      BEGIN
                        NEW(LCP) ;
                        LCP^.NAME := ID ;
                        LCP^.IDTYPE := NIL ;
                        LCP^.KLASS := FIELD ;
                        LCP^.NEXT := NIL ;
                        LCP^.VADDR := DISPL ;
                        ENTERID(LCP) ;
                        INSYMBOL ;
                        INTEST(COLON,5) ;
                        IF SY = IDENT
                          THEN
                            BEGIN
                              SEARCHID([TYPES],LCP1) ;
                              LSP1 := LCP1^.IDTYPE ;
                              IF LSP1 <> NIL
                                THEN
                                  BEGIN
                                    DISPL := DISPL + LSP1^.SIZE ;
                                    IF(LSP1^.FORM <= SUBRANGE) OR
                                      (ISSTRING(LSP1))
                                      THEN
                                        BEGIN
                                          IF ISSTRING(LSP1)
                                            THEN
                                              ERROR(399) ;
                                          LCP^.IDTYPE := LSP1
                                        END
                                      ELSE
                                        ERROR(110)
                                  END ;
                              INSYMBOL
                            END
                          ELSE
                            BEGIN
                              ERROR(2) ;
                              SKIP(FSYS +[OFSY,LPARENT])
                            END
                      END
                    ELSE
                      BEGIN
                        INSYMBOL ;
                        LSP1 := LCP1^.IDTYPE
                      END
                 END
              ELSE
                BEGIN
                  ERROR(2) ;
                  SKIP(FSYS+[OFSY,LPARENT])
                END ;
            INTEST(OFSY,8) ;
            MINSIZE := DISPL ;
            MAXSIZE := DISPL ;
            REPEAT
              LSP2 := NIL ;
              REPEAT
                CONSTANT(FSYS +[COMMA,COLON,LPARENT],LSP3,LVALU) ;
                IF NOT COMPTYPES(LSP1,LSP3)
                  THEN
                    ERROR(111) ;
                TEST := SY <> COMMA ;
                IF NOT TEST
                  THEN
                    INSYMBOL
              UNTIL TEST ;
              INTEST(COLON,5) ;
              INTEST(LPARENT,9) ;
              FIELDLIST(FSYS +[RPARENT,SEMICOLON]) ;
              IF DISPL > MAXSIZE
                THEN
                  MAXSIZE := DISPL ;
              IF SY = RPARENT
                THEN
                  BEGIN
                    INSYMBOL ;
                    TEST1(FSYS +[SEMICOLON],6)
                  END
                ELSE
                  ERROR(4) ;
              TEST := SY <> SEMICOLON ;
              IF NOT TEST
                THEN
                  BEGIN
                    DISPL := MINSIZE ;
                    INSYMBOL
                  END
            UNTIL TEST ;
            DISPL := MAXSIZE ;
          END (* CASE SYMBOL *)
      END ;(* FIELDLIST *)

    BEGIN(* TYP *)

   (* SYNTAX DIAGRAM PARSED:

              ------------
 ------------! SIMPLETYPE !-------------->------------------------------>-
    !         ------------                                             !
    V                                                                  ^
    !                     ---                        ------------      !
    !--------------------(^)----------->----------! TYPE-IDENT !->---!
    !                     ---                        ------------      !
    V  --->------                  ------<-------                      ^
    ! !          !                !              !                     !
    ! !  ------  !  -----    ---  !  ----------  !  ---    --    ----  !
    !---(PACKED)---(ARRAY)--([)---!SIMPLETYPE!---(])--(OF)--!TYPE!-!
    !    ------     -----    ---     ----------     ---    --    ----  !
    V                                                                  ^
    !              ------           ------------                       !
    !-------------(RECORD)---------! FIELD-LIST !--------->------------!
    !              ------           ------------                       !
    V                                                                  ^
    !              ------           ------------                       !
     -------------(SET )---------! SIMPLETYPE !--------->------------
                   ------           ------------                         *)
      TEST2(TYPEBEGSYS,10,FSYS) ;
      IF SY IN TYPEBEGSYS
        THEN
          BEGIN
            IF SY IN SIMPTYPEBEGSYS
              THEN
                SIMPLETYPE(FSYS,FSP,FSIZE)
            ELSE IF SY = ARROW
              THEN
                BEGIN
                 (* POINTER ^ *)
                  NEW(LSP) ;
                  FSP := LSP ;
                  LSP^.STYPE := NIL ;
                  LSP^.SIZE  := PTRSIZE ;
                  LSP^.FORM  := POINTER ;
                  INSYMBOL ;
                  IF SY = IDENT
                    THEN
                      BEGIN
                       (* SUPPRESS ERROR MESSAGE *)
                        PRTERR := FALSE ;
                        SEARCHID([TYPES],LCP) ;
                        PRTERR := TRUE ;
                        IF LCP = NIL
                          THEN
                           (* FORWARD REFERENCED TYPE ID *)
                            BEGIN
                              NEW(LCP) ;
                              LCP^.NAME := ID ;
                              LCP^.IDTYPE := LSP ;
                              LCP^.NEXT := FWPTR ;
                              LCP^.KLASS := TYPES ;
                              FWPTR := LCP
                             END
                           ELSE
			     BEGIN
                               IF LCP^.IDTYPE <> NIL
                                 THEN
			           LSP^.STYPE := LCP^.IDTYPE
			     END ;
                        INSYMBOL ;
                      END
                    ELSE
                      ERROR(2) ;
                 END
            ELSE
                BEGIN
                  IF SY = PACKEDSY
                    THEN
                      BEGIN
                       (* SKIP PACKED SYMBOL *)
                        INSYMBOL ;
                        TEST2(TYPEDELS,10,FSYS)
                      END ;
                  CASE SY OF
                     ARRAYSY :        (* ARRAY *)
                       BEGIN
                         INSYMBOL ;
                        (* '[' EXPECTED NEXT *)
                         INTEST(LBRACK,11) ;
                         LSP1 := NIL ;
                         REPEAT
                           NEW(LSP) ;
                           LSP^.STYPE := LSP1 ;
                           LSP^.INXTYPE := NIL ;
                           LSP^.FORM := ARRAYS ;
                           LSP1 := LSP ;
                           SIMPLETYPE(FSYS +[COMMA,RBRACK,OFSY],LSP2,LSIZE) ;
                           LSP1^.SIZE := LSIZE ;
                           IF LSP2 <> NIL
                             THEN
                               IF LSP2^.FORM <= SUBRANGE
                                 THEN
                                   BEGIN
                                     IF LSP2 = INTPTR
                                       THEN
                                         BEGIN
                                          (* INXTYPE MUST NOT
                                              BE INTEGER *)
                                           ERROR(149) ;
                                           LSP2 := NIL
                                          END ;
                                     LSP^.INXTYPE := LSP2
                                   END
                                 ELSE
                                   BEGIN
                                    (* INDEXTYPE MUST BE SCALAR
                                        OR SUBRANGE *)
                                     ERROR(113) ;
                                     LSP2 := NIL
                                   END ;
                               TEST := SY <> COMMA ;
                               IF NOT TEST
                                 THEN
                                   INSYMBOL
                         UNTIL TEST ;
                        (* NEXT '] OF ' EXPECTED *)
                         INTEST(RBRACK,12) ;
                         INTEST(OFSY,8) ;
                         TYP(FSYS,LSP,LSIZE) ;
                        (* CALCULATE LENGTH OF ARRAY *)
                         REPEAT
                           LSP2 := LSP1^.STYPE ;
                           LSP1^.STYPE := LSP ;
                           IF LSP1^.INXTYPE <> NIL
                             THEN
                               BEGIN
                                 GETBOUNDS(LSP1^.INXTYPE,LMIN,LMAX) ;
                                 LSIZE := LSIZE *(LMAX - LMIN + 1) ;
                                 LSP1^.SIZE := LSIZE
                               END ;
                            LSP := LSP1 ;
                            LSP1 := LSP2
                          UNTIL LSP1 = NIL
                        END ;
                     RECORDSY :       (* RECORD *)
                       BEGIN
                         INSYMBOL ;
                         OLDTOP := TOP ;
                         IF TOP < DISPLIMIT
                           THEN
                             BEGIN
                               TOP := TOP + 1 ;
                               DISPLAY[TOP] := NIL
                             END
                           ELSE
                             ERROR(250) ;(* TOO MANY NESTED *)
                        (* GET ALL FIELDS OF RECORD AS TREE *)
                         DISPL := 0 ;
                         FIELDLIST(FSYS -[SEMICOLON] +[ENDSY]);
                         NEW(LSP) ;
                         LSP^.FSTFLD := DISPLAY[TOP] ;
                         LSP^.SIZE := DISPL ;
                         LSP^.FORM := RECORDS ;
                         TOP := OLDTOP ;
                         INTEST(ENDSY,13)
                       END ;
                     SETSY :          (* SET *)
                       BEGIN
                         INSYMBOL ;
                        (* SET OF *)
                         INTEST(OFSY,8) ;
                         SIMPLETYPE(FSYS,LSP1,LSIZE) ;
                         IF LSP1 <> NIL
                           THEN
                             IF LSP1^.FORM > SUBRANGE
                               THEN
                                 BEGIN
                                   (* BASE-TYPE SCALAR OR SUBRANGE *)
                                    ERROR(115) ;
                                    LSP1 := NIL ;
                                  END ;
                          NEW(LSP) ;
                          LSP^.STYPE := LSP1 ;
                          LSP^.SIZE := SETSIZE ;
                          LSP^.FORM := POWER
                       END
                    END ;(* CASE *)
                 (* RETURN DESCRIPTION *)
                  FSP := LSP
                END ;
            TEST1(FSYS,6) ;
          END
      ELSE
        FSP := NIL ;
      IF FSP = NIL
        THEN
          FSIZE := 1
        ELSE
          FSIZE := FSP^.SIZE
    END ;(* TYP *)

    PROCEDURE CONSTDECLARATION ;
   (* COMPILES CONSTANT DECLARATION PART.
       SYNTAX DIAGRAM CHECKED(RESERVED KEYWORD CONST ALREADY PARSED):

             --------------------<---------------------------------
            !                                                      !
            V                                                      ^
            !  -------          ---      ----------         ---    !
 CONST --->---! IDENT !--------(=)----! CONSTANT !--->---(;)----->
               -------          ---      ----------         ---    *)
    VAR
      LCP : CTP;
      LSP : STP ;
      LVALU : VALU ;

    BEGIN
      TEST2([IDENT],2,FSYS) ;
      WHILE SY = IDENT  DO
        BEGIN
          NEW(LCP) ;
          LCP^.NAME := ID ;
          LCP^.IDTYPE := NIL ;
          LCP^.NEXT := NIL ;
          LCP^.KLASS := KONST ;
          INSYMBOL ;
          IF(SY = RELOP) AND(OP = EQOP)
            THEN
              INSYMBOL
            ELSE
              ERROR(16) ;
          CONSTANT(FSYS +[SEMICOLON],LSP,LVALU) ;
          ENTERID(LCP) ;
          LCP^.IDTYPE := LSP ;
          LCP^.VALUES := LVALU ;
          IF SY = SEMICOLON
            THEN
              BEGIN
                INSYMBOL ;
                TEST1(FSYS +[IDENT],6)
              END
            ELSE
              ERROR(14)
        END
    END ;(* CONSTDECLARATION *)

    PROCEDURE TYPEDECLARATION ;
   (* COMPILES TYPE-DECLARATION
       SYNTAX-DIAGRAM CHECKED(RESERVED KEYWORD TYPE ALREADY PARSED):


           ------------------------<----------------------------
          V                                                     ^
          !    -------           ---      ------------      ---  !
   TYPE ------! IDENT !---------(=)----! TYPE-IDENT !----(;)--->-
           !   -------        !  ---      ------------      ---
           ^                  V
           !  -----    ---    !
            -!IDENT!--(,)---
              -----    ---                                      *)
    VAR
      LCP,LCP1,LCP2 : CTP;
      LSP : STP ;
      LSIZE : ADDRRANGE ;

    BEGIN
      TEST2([IDENT],2,FSYS) ;
      WHILE SY = IDENT DO
        BEGIN
          NEW(LCP) ;
          LCP^.NAME := ID ;
          LCP^.IDTYPE := NIL ;
          LCP^.KLASS := TYPES ;
          INSYMBOL ;
          IF(SY = RELOP) AND(OP = EQOP)
            THEN
              INSYMBOL
            ELSE
              ERROR(16) ;
          TYP(FSYS +[SEMICOLON],LSP,LSIZE) ;
          ENTERID(LCP) ;
          LCP^.IDTYPE := LSP ;
         (* HAS ANY FORWARD REFERENCE BEEN SATISFIED *)
          LCP1 := FWPTR ;
          WHILE LCP1 <> NIL DO
            BEGIN
              IF LCP1^.NAME = LCP^.NAME
                THEN
                  BEGIN
                    LCP1^.IDTYPE^.STYPE := LCP^.IDTYPE ;
                    IF LCP1 <> FWPTR
                      THEN
                        LCP2^.NEXT := LCP1^.NEXT
                      ELSE
                        FWPTR := LCP1^.NEXT
                  END ;
               LCP2 := LCP1 ;
               LCP1 := LCP1^.NEXT
             END ;
           IF SY = SEMICOLON
             THEN
               BEGIN
                 INSYMBOL ;
                 TEST1(FSYS +[IDENT],6)
               END
             ELSE
               ERROR(14)
        END ;(* WHILE *)
      IF FWPTR <> NIL
        THEN
          BEGIN
            ERROR(117) ;
            WRITELN ;
            REPEAT
              WRITELN(' TYPE-ID ',FWPTR^.NAME) ;
              FWPTR := FWPTR^.NEXT
            UNTIL FWPTR = NIL ;
          END
    END ;(* TYPEDECLARATION *)

    PROCEDURE VARDECLARATION ;
   (* COMPILES TYPE-DECLARATION
       SYNTAX-DIAGRAM CHECKED(RESERVED KEYWORD VAR ALREADY PARSED):

           ------------------------<----------------------------
          V                                                     ^
          !   -------           ---      ------------      ---  !
   VAR ------! IDENT !---------(:)----! TYPE-IDENT !----(;)--->-
           !  -------        !  ---      ------------      ---
           ^                 V
           !  -----    ---   !
            -!IDENT!--(,)--
              -----    ---                                      *)
    VAR
      LCP,NXT : CTP;
      LSP : STP ;
      LSIZE : ADDRRANGE ;

    BEGIN
      NXT := NIL ;
      REPEAT
        REPEAT
          IF SY = IDENT
            THEN
              BEGIN
                NEW(LCP) ;
                LCP^.NAME := ID ;
                LCP^.NEXT := NXT ;
                LCP^.KLASS := VARS ;
                LCP^.IDTYPE := NIL ;
                LCP^.VKIND := ACTUAL ;
                LCP^.VLEV := LEVEL ;
                ENTERID(LCP) ;
                NXT := LCP ;
                INSYMBOL
              END
            ELSE
              ERROR(2) ;
          TEST2(FSYS + TYPEDELS +[COMMA,COLON],6,[SEMICOLON]) ;
          TEST := SY <> COMMA ;
          IF NOT TEST
            THEN
              INSYMBOL ;
        UNTIL TEST ;
        INTEST(COLON,5) ;
        TYP(FSYS +[SEMICOLON] + TYPEDELS,LSP,LSIZE) ;
        WHILE NXT <> NIL DO
          BEGIN
            NXT^.IDTYPE := LSP ;
            LC := LC + LSIZE ;
            NXT^.VADDR := LC ;
            NXT := NXT^.NEXT
          END ;
        IF SY = SEMICOLON
          THEN
            BEGIN
              INSYMBOL ;
              TEST1(FSYS +[IDENT],6)
            END
          ELSE
            ERROR(14)
      UNTIL(SY <> IDENT) AND(NOT(SY IN TYPEDELS)) ;
      IF FWPTR <> NIL
        THEN
          BEGIN
            ERROR(118) ;
            WRITELN ;
            REPEAT
              WRITELN( 'TYPE-ID ',FWPTR^.NAME) ;
              FWPTR := FWPTR^.NEXT ;
            UNTIL FWPTR = NIL ;
          END
    END ;(* VARDECLARATION *)

    PROCEDURE PROCDECLARATION(FSY : SYMBOLN) ;

    VAR
      OLDLEV   : 0 .. MAXLEVEL ;
      LSY      : SYMBOLN ;
      LCP,LCP1 : CTP;
      LSP      : STP ;
      FORW     : BOOLEAN ;
      OLDTOP   : DISPRANGE ;
      LLC,LCM  : ADDRRANGE ;
      MARKP    : MARKPT ;       (* USED FOR MARK/RELEASE *)

    PROCEDURE PARAMETERLIST(FSY  : SETOFSYS ; VAR FPAR : CTP) ;
    (* COMPILES PARAMETER-LIST IN A PROCEDURE OR FUNCTION
        DECLARATION.
        ALL VARIABLES FOUND ARE ENTERED IN THE SYMBOL-TREE
        AND THE NXT^-FIELD IS USED TO LINK THEM TOGETHER. *)

      VAR
        LCP,LCP1,LCP2,LCP3  : CTP;
        LSP     : STP ;
        LKIND   : IDKIND ;
        LLC,LEN : ADDRRANGE ;
        COUNT,
        OFFSET  : INTEGER ;

      BEGIN
        LCP1 := NIL ;
        TEST1(FSY +[LPARENT],7) ;
        IF SY = LPARENT
          THEN
            BEGIN
              IF FORW
                THEN
                  ERROR(119) ;
              INSYMBOL ;
              IF (NOT(SY IN[IDENT,VARSY]))
                THEN
                  BEGIN
                    ERROR(7) ;
                    SKIP(FSYS +[IDENT,RPARENT])
                  END ;
              WHILE SY IN[IDENT,VARSY] DO
                BEGIN
                  IF SY = VARSY
                    THEN
                      BEGIN
                        LKIND := FORMAL ;
                        INSYMBOL
                      END
                    ELSE
                      LKIND := ACTUAL ;
                  LCP2 := NIL ;
                  COUNT := 0 ;
                 (* ENTER VARIABLES IN PARAMTERLIST IN SYMBOL-TREE
                     AND LINK VAR OF SAME TYPE WITH LCP.NEXT *)
                  REPEAT
                    IF SY = IDENT
                      THEN
                        BEGIN
                          NEW(LCP) ;
                          LCP^.NAME := ID ;
                          LCP^.IDTYPE := NIL ;
                          LCP^.KLASS := VARS ;
                          LCP^.VKIND := LKIND ;
                         (* LINK VAR OF SAME TYPE *)
                          LCP^.NEXT := LCP2 ;
                          LCP^.VLEV := LEVEL ;
                          ENTERID(LCP) ;
                          LCP2 := LCP ;
                          COUNT := COUNT + 1 ;
                          INSYMBOL
                        END ;
                    IF NOT(SY IN[COMMA,COLON] + FSYS)
                      THEN
                        BEGIN
                          ERROR(7) ;
                          SKIP(FSYS +[COMMA,SEMICOLON,RPARENT])
                        END ;
                    TEST := SY <> COMMA ;
                    IF NOT TEST
                      THEN
                        INSYMBOL
                  UNTIL TEST ;
                  IF SY = COLON
                    THEN
                      BEGIN
                        INSYMBOL ;
                        IF SY = IDENT
                          THEN
                            BEGIN
                              SEARCHID([TYPES],LCP) ;
                              LSP := LCP^.IDTYPE ;
                              LCP3 := LCP2 ;
                              OFFSET := 0 ;
                              LEN := PTRSIZE ;
                              IF(LKIND = ACTUAL) AND
                                (LSP^.FORM < ARRAYS)
                                THEN
                                  IF LSP^.SIZE = 1
                                    THEN
                                      OFFSET := 0
                                    ELSE
                                      LEN := LSP^.SIZE ;
                              LC :=LC + COUNT * LEN ;
                              LLC := LC ;
                              WHILE LCP2 <> NIL DO
                                BEGIN
                                  LCP := LCP2 ;
                                  LCP2^.IDTYPE := LSP ;
                                  LCP2^.VADDR := LLC - OFFSET ;
                                  LLC := LLC- LEN ;
                                  LCP2 := LCP2^.NEXT
                                END ;
                              LCP^.NEXT := LCP1 ;
                              LCP1 := LCP3 ;
                              INSYMBOL
                            END
                          ELSE
                            ERROR(2) ;
                        TEST1(FSYS +[SEMICOLON,RPARENT],7)
                      END
                    ELSE
                      ERROR(5) ;
                  IF SY = SEMICOLON
                    THEN
                      BEGIN
                        INSYMBOL ;
                        IF NOT (SY IN[IDENT,VARSY])
                          THEN
                            BEGIN
                              ERROR(7) ;
                              SKIP(FSYS +[IDENT,RPARENT])
                            END
                      END
                END ;(* WHILE *)
              IF SY = RPARENT
                THEN
                  BEGIN
                    INSYMBOL ;
                    TEST1(FSY + FSYS,6)
                  END
                ELSE
                  ERROR(4) ;
              LCP3 := NIL ;
             (* REVERSE POINTERS AND RESERVE LOCAL CELLS
                 FOR COPIES OF MULTIPLE VALUES  *)
              WHILE LCP1 <> NIL DO
                BEGIN
                  LCP2 := LCP1^.NEXT ;
                  LCP1^.NEXT := LCP3 ;
                  IF LCP1^.KLASS = VARS
                    THEN
                      IF LCP1^.IDTYPE <> NIL
                        THEN
                          IF(LCP1^.VKIND = ACTUAL) AND
                            (LCP1^.IDTYPE^.FORM > POWER)
                            THEN
                              BEGIN
                                LC := LC + LCP1^.IDTYPE^.SIZE ;
                                LCP1^.VADDR := LC
                              END ;
                  LCP3 := LCP1 ;
                  LCP1 := LCP2
                END ;(* WHILE *)
              FPAR := LCP3
            END
          ELSE
           FPAR := NIL
 			;
      END ;(* PARAMETERLIST *)

     (* PROCDECLARATION *)
     (* PROCEDURE TO COMPILE A PROCEDURE OR A FUNCTION
         DECLARATION.
         NOTE THAT SYMBOLS PROCSY OR FUNCSY ALREADY PARSED.
         ENTERED WITH CURRENT SYMBOL IS PROC/FUNC IDENTIFIER.
         SYNTAX DIAGRAM PARSED :

    --------    -------    -    ----------     -
->-!FUNCT-ID!--!PARLIST!--(:)--!TYPE-IDENT!---(;)------
    --------    -------    -    ----------     -  ^    !
                                                  !    V
    --------    -------    -                      !    !
->-!PROC-ID !--!PARLIST!--(;)---------------------     !
    --------    -------    -                           !
                                                       !
   -----------------------<----------------------------
  !
  V
  !        -------             -
   --->---! BLOCK !-----------(;)------>----------->-
  !        -------             -           !
  V                                        ^
  !        ---------                       !
   --->---(FORWARD)------------------>---
  !        ---------                       !
  V                                        ^
  !        --------    -    --------       !
   --->---(EXTERN)--(=)--! NUMBER !-->---
           --------    -    --------
                                                         *)
     BEGIN
      LLC := LC ;
      LC := LCAFTERMARKSTACK ;
      IF SY = IDENT
        THEN
          BEGIN
           (* DECIDE IF FORWARD BY SEARCHING FOR PROC/FUNC
               IDENTIFIER IN SYMBOLTABLE AT CURRENT LEVEL.
               IF NOT FOUND THERE IT CAN NOT BE FORWARD *)
            SEARCHSECTION(DISPLAY[TOP],LCP) ;
            IF LCP <> NIL
              THEN
                BEGIN
                  IF LCP^.KLASS = PROC
                    THEN
                      FORW :=(LCP^.FORWDECL) AND(FSY = PROCSY)
                  ELSE IF LCP^.KLASS = FUNC
                    THEN
                      FORW :=(LCP^.FORWDECL) AND(FSY = FUNCSY)
                  ELSE
                    FORW := FALSE ;
                  IF NOT FORW
                    THEN
                      ERROR(160)
                     (* PREVIOUS DECLARATION WAS NOT FORWARD *)
                END
              ELSE
                FORW := FALSE ;
            IF NOT FORW
              THEN
               (* NEW PROCEDURE OR FUNCTION DESCRIPTION *)
                BEGIN
                  NEW(LCP) ;
                  LCP^.NAME := ID ;
                  LCP^.IDTYPE := NIL ;
                  LCP^.PFLEV := LEVEL ;
                  LCP^.PFNAME := NPROC ;
                  NPROC := NPROC + 1 ;
                  LCP^.EXTERNL := FALSE ;
                  LCP^.PFDECKIND := DECLARED ;
                  IF FSY = PROCSY
                    THEN
                      LCP^.KLASS := PROC
                    ELSE
                      LCP^.KLASS := FUNC ;
                 (* ENTER DESCRIPTION AT CURRENT LEVEL *)
                  ENTERID(LCP)
                END
              ELSE
                BEGIN
                 (* DESCRIPTION OF PROC/FUNC ALREADY PRESENT
                     INCLUDING FIELDLIST AS LINKED LIST POINTED
                     AT BY NEXT^ IN FORWARD DECLARATION     *)
                 (* WALK THROUGH LINKED LIST FOR SPACE TO RESERVE *)
                  LCP1 := LCP^.NEXT ;
                  WHILE LCP1 <> NIL DO
                    BEGIN
                      IF LCP1^.KLASS = VARS
                        THEN
                          IF LCP1^.IDTYPE <> NIL
                            THEN
                              BEGIN
                                LCM := LCP1^.VADDR ;
                                IF LCM > LC
                                  THEN
                                    LC := LCM
                              END ;
                         (* NEXT FIELD IN LINKED LIST *)
                          LCP1 := LCP1^.NEXT
                    END ;
                END ;
            INSYMBOL
          END  (* SYMBOL IS IDENT *)
        ELSE
         (* IDENTIFIER OF PROC/FUNC EXPECTED *)
          ERROR(2) ;
     (* REMEMBER CURRENT LEVEL *)
      OLDLEV := LEVEL ;
      OLDTOP := TOP ;
      IF LEVEL < MAXLEVEL
        THEN
          LEVEL := LEVEL + 1
        ELSE
          ERROR(251) ;
     (* REMEMBER CURRENT DISPLAY *)
      IF TOP < DISPLIMIT
        THEN
          BEGIN
            TOP := TOP + 1 ;
            SAVETOP := TOP ;
            IF FORW
              THEN
               (* ALREADY FIELDLIST PRESENT IN TREE *)
                DISPLAY[TOP] := LCP^.NEXT
              ELSE
               (* CREATE EMPTY TREE *)
                DISPLAY[TOP] := NIL
          END
        ELSE
          ERROR(1250) ;
      IF FSY = PROCSY
        THEN
          BEGIN
            PARAMETERLIST([SEMICOLON],LCP1) ;
            IF NOT FORW
              THEN
               (* LINK LINKED LIST TO PROC DESCRIPTION *)
                LCP^.NEXT := LCP1
          END
        ELSE
         (* FUNCTION *)
          BEGIN
            PARAMETERLIST([SEMICOLON,COLON],LCP1) ;
            IF NOT FORW
              THEN
               (* LINK LINKED LIST TO FUNC DESCRIPTION *)
                LCP^.NEXT := LCP1 ;
            IF SY = COLON
              THEN
                BEGIN
                 (* DETERMINE TYPE OF FUNCTION-RESULT *)
                  INSYMBOL ;
                  IF SY = IDENT
                    THEN
                      BEGIN
                        IF FORW
                          THEN
                           (* FUNCTION-TYPE AGAIN DECLARED *)
                            ERROR(122) ;
                        SEARCHID([TYPES],LCP1) ;
                        LSP := LCP1^.IDTYPE ;
                        LCP^.IDTYPE := LSP ;
                        IF LSP <> NIL
                          THEN
                            IF LSP^.FORM >= POWER
                              THEN
                                BEGIN
                                 (* FUNCTION-TYPE MUST BE SIMPLE *)
                                  ERROR(120) ;
                                  LCP^.IDTYPE := NIL
                                END ;
                        INSYMBOL
                      END
                    ELSE
                      BEGIN
                       (* IDENTIFIER EXPECTED *)
                        ERROR(2) ;
                        SKIP(FSYS +[SEMICOLON])
                      END
                END
              ELSE
                IF NOT FORW
                    THEN
                     (* COLON EXPECTED WITH FUNCTION-TYPE*)
                      ERROR(123)
          END ;
      INTEST(SEMICOLON,14) ;
      IF SY = FORWARDSY
        THEN
          BEGIN
            IF FORW
              THEN
                ERROR(161)
              ELSE
                LCP^.FORWDECL := TRUE ;
            INSYMBOL ;
            INTEST(SEMICOLON,14) ;
            TEST1(FSYS,6)
          END
      ELSE IF SY = EXTERNSY
        THEN
          BEGIN
            IF FORW
              THEN ERROR(161) ;
            INSYMBOL ;
            LCP^.EXTERNL := TRUE ;
            IF(SY = RELOP) AND(OP = EQOP)
              THEN
                BEGIN
                  INSYMBOL ;
                  IF SY = INTCONST
                    THEN
                      LCP^.PFNAME := VAL.IVAL
                    ELSE
                      ERROR(106)
                END
              ELSE
                 ERROR(106) ;
            REPEAT
              INSYMBOL
            UNTIL SY = SEMICOLON ;
            INSYMBOL ;
            TEST1(FSYS,6)
          END
        ELSE
         (* BLOCK *)
          BEGIN
            LCP^.FORWDECL := FALSE ;
           (* REMEMBER HEAP WITH MARKPOINTER FOR RELEASE *)
            NEW(MARKP) ;
            REPEAT
              BLOCK(FSYS,SEMICOLON,LCP) ;
              IF SY = SEMICOLON
                THEN
                  BEGIN
                    INSYMBOL ;
                    IF NOT(SY IN[BEGINSY,PROCSY,FUNCSY])
                      THEN
                        BEGIN
                          ERROR(6) ;
                          SKIP(FSYS)
                        END
                  END
                ELSE
                  ERROR(14) ;
            UNTIL SY IN[BEGINSY,PROCSY,FUNCSY] ;
           (* RELEASE HEAP FROM MARKPOINTER *)
            RELEASE(MARKP) ;
          END ;
       (* RESTORE TO OLD LEVEL AND DISPLAY *)
        LEVEL := OLDLEV ;
        TOP := OLDTOP ;
        LC := LLC
    END ;(* PROCDECLARATION *)

    PROCEDURE BODY(FSYS : SETOFSYS) ;

    VAR
      I,ENTNAME,SEGSIZE : INTEGER ;
      LCMAX,LLC1 : ADDRRANGE ;
      LCP : CTP;

      PROCEDURE LDAGEN (LEVEL : INTEGER ; DPLMT : ADDRRANGE) ;
     (* GENERATE INSTRUCTIONS TO LOAD OFFSET ADDRESS DPLMT
         FOR LEVEL ON STACK,OPTIMIZE IF SHORT RANGE *)
        BEGIN
          IF DPLMT < 256
            THEN                          (* LDAS *)
              BEGIN
                BYTEGEN(16  + LEVEL) ;
                BYTEGEN(DPLMT)
              END
            ELSE                          (* LDA  *)
              BEGIN
                BYTEGEN(32 + LEVEL) ;
                ADDRGEN(DPLMT)
              END
        END ;(* LDAGEN *)

      PROCEDURE LODGEN(LEVEL : INTEGER ; VAR FATTR : ATTR) ;
     (* LOAD VARIABLE VALUE ON STACK OF LEVEL *)

        BEGIN
         (* SHORT RANGE AND SMALL VARIABLE *)
          IF(FATTR.TYPTR^.SIZE <= 2) AND (FATTR.DPLMT < 256)
            THEN
              BEGIN
               (* LOD1,LOD2 OR LOD8 BYTE INSTRUCTIONS *)
                BYTEGEN(64 + 16 * FATTR.TYPTR^.SIZE + LEVEL) ;
                BYTEGEN(FATTR.DPLMT)
              END
            ELSE
              BEGIN
               (* PUT ADDRESS ON STACK OF DATA ITEM FROM LEVEL *)
                LDAGEN(LEVEL,FATTR.DPLMT) ;
               (* INDIRECTLY LOAD DATA ON STACK *)
                CASE FATTR.TYPTR^.SIZE OF
                  1 : BYTEGEN(154) ; (* IND1 *)
                  2 : BYTEGEN(155) ; (* IND2 *)
                  3 : BYTEGEN(156)   (* IND8 *)
                END
              END
        END ;(* LODGEN *)

      PROCEDURE CONDGEN(FOP : OPRANGE ; VAR FATTR : ATTR) ;
     (* GENERATE CONDITIONAL OPERERATORS,
         CALLED WITH OPERATORS FOR 2-BYTE ITEMS,
         FOR OTHER ITEMS CALCULATE OPERATOR-CODE *)

        BEGIN
          IF FATTR.TYPTR^.FORM > POWER
            THEN
              BEGIN
               (* ARRAYS AND RECORDS *)
               (* LEQ2(= $90) TO LEQM(= $92) *)
               (* LES2(= $93) TO LESM(= $95) *)
               (* EQU2(= $96) TO EQUM(= $98) *)
                BYTEGEN(FOP + 2) ;
                ADDRGEN(FATTR.TYPTR^.SIZE)
              END
          ELSE IF FATTR.TYPTR^.SIZE = 1
            THEN
               (* CORRECT TO 16 BIT *)
               (* LEQ2(= $90) *)
               (* LES2(= $93) *)
               (* EQU2(= $96) *)
              BYTEGEN(FOP)
          ELSE IF FATTR.TYPTR^.SIZE = 2
            THEN
               (* LEQ2(= $90) *)
               (* LES2(= $93) *)
               (* EQU2(= $96) *)
              BYTEGEN(FOP)
          ELSE IF FATTR.TYPTR^.SIZE = 8
            THEN
               (* SETS,ONLY EQUAL OPERATOR *)
               (* EQU2(= $96) TO EQUM(= $99) *)
              BYTEGEN(FOP + 3)
          ELSE
            ERROR(401)          (* WRONG SIZE *)
        END ;(* CONDGEN *)

      PROCEDURE LDCIGEN(VALUE : INTEGER) ;
       (* GENERATE INSTRUCTIONS TO LOAD CONSTANT
           16 BIT VALUE ON STACK *)

        BEGIN
          IF(VALUE < 16) AND(VALUE >= 0)
            THEN
              BYTEGEN(VALUE)
          ELSE IF VALUE < 0
            THEN
              BEGIN
                BYTEGEN(176) ;              (* LNC *)
                ADDRGEN(-VALUE)
              END
          ELSE
            BEGIN
              BYTEGEN(160) ;                (* LDC *)
              ADDRGEN(VALUE)
            END
        END ;(* LDCIGEN *)

      PROCEDURE LDCS(S : INTSET) ;
     (* GENERATE INSTRUCTION TO
         LOAD CONSTANT SET S ON STACK *)

        VAR
          I,K,L,N : INTEGER ;
          B : BOOLEAN ;

        BEGIN
          N := 0 ;
          B := FALSE ;
          FOR I := 0 TO SETMAX DO
            IF I IN S
              THEN
                N := N + 1 ;
          IF N = 0
            THEN
              BYTEGEN(193)           (* LOAD EMPTY SET LNS *)
          ELSE IF N < 4
            THEN
              BEGIN
                FOR I := 0 TO SETMAX DO
                  IF I IN S
                    THEN
                      BEGIN
                        LDCIGEN(I) ;
                        BYTEGEN(174) ;     (* SGS *)
                        IF B
                          THEN
                            BYTEGEN(175) ; (* UNI *)
                        B := TRUE
                      END
               END
          ELSE        (* N >= 4 *)
            BEGIN
              BYTEGEN(186) ;               (* LDCS *)
              L := 0 ;
              K := 128 ;
              FOR I := 0 TO SETMAX DO
                BEGIN
                  IF K < 1
                    THEN
                      BEGIN
                        K := 128 ;
                        BYTEGEN(L) ;
                        L := 0
                      END ;
                  IF I IN S
                    THEN
                      L := L + K ;
                 (* SHIFT *)
                  K := K DIV 2
                END ;
              BYTEGEN(L)
            END
        END ;(* LDCS *)

      PROCEDURE CSPGEN(P : INTEGER) ;
     (* GENERATE INSTRUCTION TO CALL STANDARD PROCEDURE STPR P *)

        BEGIN
          BYTEGEN(189 ) ;
          BYTEGEN(P)
        END ;(* CSPGEN *)


      PROCEDURE GENINC(INC : INTEGER) ;
     (* GENERATE INSTRUCTIONS TO INCREMENT OR
         DECREMENT INTEGER ON STACK WITH INC *)

        BEGIN
          IF INC <> 0
            THEN
              IF INC = 1
                THEN
                  BYTEGEN(185)                (* INC1 *)
              ELSE IF INC = -1
                THEN
                  BYTEGEN(184)                (* DEC1 *)
              ELSE
                IF INC > 0
                  THEN
                    BEGIN
                      BYTEGEN(180) ;          (* INC *)
                      ADDRGEN(INC)
                    END
                  ELSE
                    BEGIN
                      BYTEGEN(179) ;          (* DEC *)
                      ADDRGEN(- INC)
                    END
        END ;(* GENINC *)

      PROCEDURE LOAD ;
     (* GENERATE INSTRUCTION TO LOAD DATA ON STACK *)

      BEGIN
        IF GATTR.TYPTR <>  NIL
          THEN
            BEGIN
              (* CONSTANT *)
              IF GATTR.KIND = CST
                THEN
                  BEGIN
                    IF(GATTR.TYPTR^.FORM = SCALAR)
                      THEN
                       (* LOAD CONSTANT ON STACK *)
                        LDCIGEN(GATTR.CVAL.IVAL)
                    ELSE IF GATTR.TYPTR = NILPTR
                      THEN
                       (* LOAD ZERO ON STACK *)
                       BYTEGEN(0)
                    ELSE
                      (* LOAD CONSTANT SET ON STACK *)
                      LDCS(GATTR.CVAL.VALP^.PVAL)
				  END
              ELSE IF GATTR.KIND = VARBL
                THEN
                 IF GATTR.ACCESS = DRCT
                  THEN
                    LODGEN(LEVEL - GATTR.LEVEL,GATTR)
                  ELSE
                    BEGIN
                     (* PUT ADDRESS OF DATA ITEM ON STACK *)
                      GENINC(GATTR.DPLMT) ;
                     (* INDIRECTLY LOAD DATA ITEM ON STACK *)
                      CASE GATTR.TYPTR^.SIZE OF
                        1 : BYTEGEN(154) ;   (* IND1 *)
                        2 : BYTEGEN(155) ;   (* IND2 *)
                        8 : BYTEGEN(156) ;   (* IND8 *)
                      END
                    END ;
            GATTR.KIND := EXPR
            END
      END ;(* LOAD *)

      PROCEDURE STORE(VAR FATTR : ATTR) ;
    (* GENERATE INSTRUCTION TO STORE DATA ITEM IN MEMORY *)

      BEGIN
        IF FATTR.TYPTR <> NIL
          THEN
            IF FATTR.ACCESS = DRCT
              THEN
                BEGIN
                 (* STR1 OR LOD2 *)
                  BYTEGEN(96 + 16 * FATTR.TYPTR^.SIZE
                            + LEVEL - FATTR.LEVEL) ;
                  BYTEGEN(FATTR.DPLMT)
                END
              ELSE
                IF FATTR.DPLMT <> 0
                  THEN
                    ERROR(400)
                  ELSE
                    CASE FATTR.TYPTR^.SIZE OF
                      1 : BYTEGEN(157) ;         (* STO1 *)
                      2 : BYTEGEN(158) ;         (* STO2 *)
                      8 : BYTEGEN(159) ;         (* STO8 *)
                    END
      END ;(* STORE *)

      PROCEDURE LOADADDRESS ;
     (* GENERATE INSTRUCTIONS TO LOAD ADDRESS OF VARIABLE
         OR CONSTANT ON STACK *)

      VAR
        LVP : CSP ;
        I : INTEGER ;

      BEGIN
        IF GATTR.TYPTR <> NIL
          THEN
            BEGIN
              IF GATTR.KIND = CST
                THEN
				  BEGIN
                    IF ISSTRING(GATTR.TYPTR)
                      THEN
                        BEGIN
                          (* PUT ADDRESS OF STRING ON STACK *)
                          BYTEGEN(188) ;        (* LCA *)
                          BYTEGEN(STRINGSIZE(GATTR.CVAL.VALP)) ;
                          LVP := GATTR.CVAL.VALP ;
                         (* PUT STRING FOLLOWING OPCODE *)
                          REPEAT
                            FOR I := 1 TO LVP^.SLGTH DO
                              BYTEGEN(LVP^.SVAL[I]) ;
                           (* WALK THROUGH LINKED LIST OF STRING PARTS *)
                            LVP := LVP^.SPTR
                          UNTIL LVP = NIL
                        END
                      ELSE
                        ERROR(400)
                  END
              ELSE IF GATTR.KIND = VARBL
                THEN
                  BEGIN
				    IF GATTR.ACCESS = DRCT
                      THEN
                        LDAGEN(LEVEL - GATTR.LEVEL,GATTR.DPLMT)
                      ELSE
                        GENINC(GATTR.DPLMT)
				  END
              ELSE
                 ERROR(400) ;
              GATTR.KIND := VARBL ;
              (* ADDRESS ON STACK,SO MARK AS INDIRECT *)
              GATTR.ACCESS := INDRCT ;
              GATTR.DPLMT := 0
            END
      END ;(* LOADADDRESS *)

      PROCEDURE GENFJP(FADDR : INTEGER) ;
     (* GENERATE INSTRUCTION JUMP CONDITIONAL
         IF FALSE VALUE ON STACK JUMP TO FADDR *)

      BEGIN
        LOAD ;
        IF GATTR.TYPTR <> NIL
          THEN
            IF GATTR.TYPTR <> BOOLPTR
              THEN
               (* MUST HAVE BOOLEAN RESULT ON STACK *)
                ERROR(144) ;
        BYTEGEN(177)   ;          (* FJP *)
        ADDRGEN(FADDR)
      END ;(* GENFJP*)

      PROCEDURE GENCUP(FP1,FP2 : INTEGER) ;
     (* GENERATE CALL TO USER PROCEDURE *)

      BEGIN
       (* PROC/FUNC IN PARAMETER LIST ? *)
        IF PFCTTEST
          THEN
            BEGIN
              BYTEGEN(FP1) ;
              BYTEGEN(191)        (* CUP2 *)
            END
          ELSE
            BYTEGEN(190) ;        (* CUP1 *)
        BYTEGEN(FP2)
      END ;(* GENCUP *)

      PROCEDURE STATEMENT(FSYS : SETOFSYS) ;

      VAR LCP : CTP;

        PROCEDURE EXPRESSION(FSYS : SETOFSYS) ; FORWARD ;

          PROCEDURE SELECTOR(FSYS : SETOFSYS ; FCP : CTP) ;
          (* DEALS WITH COMPLEX IDENTIFIERS SUCH AS POINTERS,
             FIELDS OF RECORDS AND INDEXES IN ARRAYS AND
             COMBINATIONS,RESULT RETURNED IN GATTR         *)
             (* STRUCTURES PARSED,FIRST IDENTIFIER ALREADY PARSED :
                                    ---
                      -------------(^)----------------
                     !              ---                 !
                     ^                                  !
      ------------   !   ---     ------------     ---   V
->---! IDENTIFIER !--!--(()---! EXPRESSION !---() )------>
   ^  ------------   !   ---  ^  ------------  !  ---   ^  !
   !                 !        !                !        !  V
   !                 V         -------<--------         !  !
   !                 !                                  !  !
   !                 !               ---                !  !
   !                  --------------(.)---------------   !
   !                                 ---                   !
   !                                                       !
    ----------------------------------<--------------------   *)
          VAR
            LATTR : ATTR ;
            LCP : CTP;
            LMIN,LMAX : INTEGER ;

        BEGIN
          GATTR.TYPTR := FCP^.IDTYPE ;
          GATTR.KIND := VARBL ;
          IF FCP^.KLASS = VARS
            THEN
              BEGIN
               (* DECIDE IF DIRECT OR INDIRECT ACCESS *)
                IF GATTR.TYPTR <> NIL
                  THEN
                    IF(FCP^.VKIND = ACTUAL) AND (FCP^.VADDR < 256) AND
					  (GATTR.TYPTR^.SIZE <= 2)
                      THEN
                        BEGIN
                          GATTR.ACCESS := DRCT ;
                          GATTR.LEVEL := FCP^.VLEV ;
                          GATTR.DPLMT := FCP^.VADDR
                        END
                      ELSE
                        BEGIN
                          GATTR.ACCESS := INDRCT ;
                          GATTR.DPLMT := 0 ;
                          IF FCP^.VKIND = FORMAL
                            THEN
                              BEGIN
                                BYTEGEN(96 + LEVEL - FCP^.VLEV) ;
                                BYTEGEN(FCP^.VADDR)
                              END
                            ELSE
                              LDAGEN(LEVEL - FCP^.VLEV,FCP^.VADDR)
                        END
              END
          ELSE IF FCP^.KLASS = FUNC
            THEN
              IF FCP^.PFDECKIND = STANDARD
                THEN
                 (* ASSIGNMENT TO STANDARD FUNCTION NOT ALLOWED *)
                  ERROR(150)
                ELSE
                  BEGIN
                   (* IMPLIED RELATIVE ADDRESS OF FUNCTION RESULT *)
                    GATTR.ACCESS := DRCT ;
                    GATTR.LEVEL := FCP^.PFLEV + 1 ;
                    GATTR.DPLMT := 0
                  END ;
         (* TEST IF NOT ERROR IN VARIABLE,MUST BE '^','(' OR '.' *)
          TEST1(SELECTSYS + FSYS,59) ;
          WHILE SY IN SELECTSYS DO
            BEGIN
			  (* [ *)
              IF SY = LBRACK
                THEN
                  BEGIN
                   (* ARRAY ELEMENT *)
                    REPEAT
                     (* ALL INDEXES SEPARATED BY COMMA'S *)
                      LATTR := GATTR ;
                     (* SAVE EXPRESSION RESULT *)
                      IF LATTR.TYPTR <> NIL
                        THEN
                          IF LATTR.TYPTR^.FORM <> ARRAYS
                            THEN
                              BEGIN
                               (* TYPE OF VARIABLE IS NOT ARRAY *)
                                ERROR(138) ;
                                LATTR.TYPTR := NIL
                              END ;
                      (* COMPLEX STRUCTURE,REMEMBER ADDRESS *)
                      LOADADDRESS ;
                      INSYMBOL ;
                     (* INDEX IS EXPRESSION FOLLOWED BY ',' OR ']' *)
                      EXPRESSION(FSYS +[COMMA,RBRACK]) ;
                      IF GATTR.KIND <> CST
                        THEN
                          LOAD ;
                      IF GATTR.TYPTR <> NIL
                        THEN
                          IF GATTR.TYPTR^.FORM > SUBRANGE
                            THEN
                             (* INDEXTYPE MUST BE SCALAR OR SUBRANGE *)
                              ERROR(113) ;
                      IF LATTR.TYPTR <> NIL
                        THEN
                          BEGIN
                           (* CHECK IF INDEX SAME TYPE AS ARRAY INDEX *)
                            IF COMPTYPES(LATTR.TYPTR^.INXTYPE, GATTR.TYPTR)
                              THEN
                                BEGIN
                                  IF LATTR.TYPTR^.INXTYPE <> NIL
                                    THEN
                                      BEGIN
                                        GETBOUNDS(LATTR.TYPTR^.INXTYPE,
                                                  LMIN,LMAX) ;
                                        IF GATTR.KIND = CST
                                          THEN
                                            BEGIN
                                             (* CHECK BOUNDS IF
                                                 CONSTANT *)
                                              IF (GATTR.CVAL.IVAL < LMIN) OR
                                               (GATTR.CVAL.IVAL > LMAX)
                                                THEN
                                                  ERROR(302) ;
                                              GATTR.CVAL.IVAL :=
                                                GATTR.CVAL.IVAL - LMIN
                                            END
                                          ELSE
                                            GENINC(-LMIN)
                                           (* DECREMENT OFFSET WITH
                                               LMIN FOR ACTUAL OFFSET *)
                                      END
                                END
                              ELSE
                                ERROR(139) ;
                            GATTR.TYPTR := LATTR.TYPTR^.STYPE ;
                            IF GATTR.TYPTR <> NIL
                              THEN
                                IF GATTR.KIND = CST
                                  THEN
                                    BEGIN
                                      GATTR.CVAL.IVAL :=
                                          GATTR.CVAL.IVAL *
                                             GATTR.TYPTR^.SIZE ;
                                      LDCIGEN(GATTR.CVAL.IVAL)
                                    END
                                  ELSE
                                    BEGIN
                                      IF GATTR.TYPTR^.SIZE <> 1
                                        THEN
                                          BEGIN
                                            LDCIGEN(GATTR.TYPTR^.SIZE);
                                            BYTEGEN(170)  (* MPI *)
                                          END
                                    END ;
                            BYTEGEN(162) ;   (* ADI *)
                            GATTR.KIND := VARBL ;
                            GATTR.ACCESS := INDRCT ;
                            GATTR.DPLMT := 0
                          END
                    UNTIL SY <> COMMA ;
                   (* INDEX-LIST ENDS WITH ']' *)
                    INTEST(RBRACK,12)
                  END  (* ARRAY-INDEX *)
              ELSE IF SY = PERIOD
                THEN
                 (* RECORD-FIELD *)
                  BEGIN
                    IF GATTR.TYPTR <> NIL
                      THEN
                        IF GATTR.TYPTR^.FORM <> RECORDS
                          THEN
                            BEGIN
                             (* NOT A RECORD *)
                              ERROR(140) ;
                              GATTR.TYPTR := NIL
                            END ;
                    INSYMBOL ;
                    IF SY = IDENT
                      THEN
                        BEGIN
                          IF GATTR.TYPTR <> NIL
                            THEN
                              BEGIN
                               (* CHECK IF KNOWN FIELD IN RECORD *)
                                SEARCHSECTION(GATTR.TYPTR^.FSTFLD,LCP) ;
                                IF LCP = NIL
                                  THEN
                                    BEGIN
                                      ERROR(152) ;
                                      GATTR.TYPTR := NIL
                                    END
                                  ELSE
                                    BEGIN
                                      GATTR.TYPTR := LCP^.IDTYPE ;
                                      IF GATTR.ACCESS = INDRCT
                                        THEN
                                          GATTR.DPLMT :=
                                             GATTR.DPLMT + LCP^.VADDR
                                         ELSE
                                           GATTR.DPLMT :=
                                            GATTR.DPLMT - LCP^.VADDR
                                     END
                              END ;
                          INSYMBOL
                        END
                      ELSE
                       ERROR(2)
                  END  (* SY = PERIOD . *)
              ELSE
               (* MUST BE ^ OF POINTER *)
                BEGIN
                  IF GATTR.TYPTR <> NIL
                    THEN
                      IF GATTR.TYPTR^.FORM = POINTER
                        THEN
                          BEGIN
                            LOAD ;
                            GATTR.TYPTR := GATTR.TYPTR^.STYPE ;
                            GATTR.KIND := VARBL ;
                            GATTR.ACCESS := INDRCT ;
                            GATTR.DPLMT := 0
                          END
                        ELSE
                          ERROR(141) ;
                  INSYMBOL
                END ;  (* SY = ARROW ^ *)
              TEST1(FSYS + SELECTSYS,6)
            END ;(* WHILE *)
        END ;(* SELECTOR *)

        PROCEDURE CALL(FSYS : SETOFSYS ; FCP : CTP) ;

        VAR
          LKEY : 1 .. MAXSTANDRD  ;

        PROCEDURE VARIABLE(FSYS : SETOFSYS) ;
       (* SEARCH IF KNOWN VARIABLE AT CURRENT OR LOWER LEVEL,
           RETURN COMPLETE DESCRIPTION BY CALLING SELECTOR *)

        VAR
          LCP : CTP;

        BEGIN
          IF SY = IDENT
            THEN
              BEGIN
               (* LOOK IN SYMBOL-TREE *)
                SEARCHID([VARS],LCP) ;
                INSYMBOL
              END
            ELSE
              BEGIN
                ERROR(2) ;
               (* RETURN UNKNOWN VARIABLE POINTER *)
                LCP := UVARPTR
              END ;
         (* GET COMPLETE DESCRIPTION *)
          SELECTOR(FSYS,LCP)
        END ;(* VARIABLE *)

        PROCEDURE READP ;
       (* GENERATE CODE FOR STANDARD PROCEDURE READ AND READLN

           PARSED STRUCTURE,READLN OR READ ALREADY PARSED IN LKEY :

              ----
            -(LN)------------------->-------------------------
           ^  ----  !                                           !
           !        !            ------->----------             !
           !        !           ^                  !            !
           !        V     ---   !  ------------    V            !
 READ -->----------------(()----(FILE-IDENT)------          !
                          ---      ------------       !         !
                                                      V         V
          -----------------<------------------------------>-----!
         V                                                      !
         !                     --------------------        ---  V
          --------------------! CHARACTER-VARIABLE !------())---->--
                        !  !   --------------------   ! !  ---
                        ^  V                          ^ V
                        !  !   --------------------   ! !
                        !   --! INTEGER-VARIABLE   !--  !
                        !      --------------------     !
                        !                               !
                         --------------<----------------          *)


        BEGIN
          IF SY = LPARENT
            THEN
              BEGIN
                INSYMBOL ;
                IF SY = IDENT
                  THEN
                    BEGIN
                      REPEAT
                        VARIABLE(FSYS +[COMMA,RPARENT]) ;
                       (* PLACE ADDRESS OF VARIABLE
                           ON STACK FOR RDC OR RDI *)
                        LOADADDRESS ;
                        IF GATTR.TYPTR <> NIL
                          THEN
                            BEGIN
                              IF GATTR.TYPTR^.FORM <= SUBRANGE
                                THEN
                                  IF COMPTYPES(INTPTR,GATTR.TYPTR)
                                    THEN
                                      CSPGEN(3)           (* RDI *)
                                    ELSE IF COMPTYPES(CHARPTR,GATTR.TYPTR)
                                      THEN
                                        CSPGEN(5)           (* RDC *)
                                    ELSE
                                      ERROR(399)
                            END
                              ELSE
                                (* ERROR IN TYPE OF STANDARD PROCEDURE *)
                                ERROR(116) ;
                        TEST := SY <> COMMA ;
                        IF NOT TEST
                          THEN
                            INSYMBOL
                      UNTIL TEST ;
                   END ;
                INTEST(RPARENT,4)
              END  ;
            IF LKEY = 5
             THEN
               CSPGEN(4)               (* RLN *)
       END ;(* READP *)

       PROCEDURE WRITEP ;
       (* GENERATE CODE FOR STANDARD PROCEDURE WRITE AND WRITELN
         PARSED STRUCTURE,WRITELN OR WRITE ALREADY PARSED IN LKEY :
              ----
            -(LN)------------------->------------------------
           ^  ----  !                                          !
           !        !            ------->----------            !
           !        !           ^                  !           !
           !        V     ---   !  ------------    V           !
 WRITE-->----------------(()----(FILE-IDENT)---------->----!
                          ---      ------------       !        !
                                                      V        V
  -------------------------<--------------------------         !
 !                                                             !
 V                                                             !
 !    -----------------                                        !
 !->-! STRING          !---                                    !
 !    -----------------    !   -------------->---------        !
 !                         V  ^                        !       !
 !    -----------------    !  !  ---    -------------  V  ---  V
 !->-! CHAR-EXPRESSION !->------(:)--! INT-EXPRESS !---())---->--
 !    -----------------    !     ---    -------------     ---
 !                         ^
 !    -----------------    !
  ->-! INT-EXPRESSION  !---
      -----------------                                            *)

       VAR
         LSP : STP ;
         DEFAULT : BOOLEAN ;
         LLKEY : 1 .. MAXSTANDRD ;
         LEN : ADDRRANGE ;

        BEGIN
          LLKEY := LKEY ;
          IF SY = LPARENT
            THEN
              BEGIN
                INSYMBOL ;
                IF SY IN FACBEGSYS
                 THEN
                   BEGIN
                     REPEAT
                       EXPRESSION(FSYS +[COMMA,COLON,RPARENT]) ;
                       LSP := GATTR.TYPTR ;
                       IF LSP <> NIL
                         THEN
                          (* PUT VARIABLE OR ADDRESS OF
                              VARIABLE ON STACK *)
                           IF LSP^.FORM <= SUBRANGE
                             THEN
                               LOAD
                             ELSE
                               LOADADDRESS ;
                      (* IF COLON THEN NO DEFAULT,LOAD
                          EXPRESSION AS NR OF CHARACTERS *)
                       IF SY = COLON
                         THEN
                           BEGIN
                             INSYMBOL ;
                             EXPRESSION(FSYS +[COMMA,RPARENT]) ;
                             IF GATTR.TYPTR <> NIL
                               THEN
                                 IF GATTR.TYPTR <> INTPTR
                                   THEN
                                      ERROR(116) ;  
                             LOAD ;
                             DEFAULT := FALSE
                           END
                         ELSE
                           DEFAULT := TRUE ;
                       IF LSP = INTPTR
                         THEN
                           BEGIN
                             IF DEFAULT
                               THEN
                                (* DEFAULT INTEGER IN 6 CHAR FIELD *)
                                 LDCIGEN(6) ;
                             CSPGEN(0)         (* WRI *)
                           END
                       ELSE IF LSP = CHARPTR
                         THEN
                           BEGIN
                             IF DEFAULT
                               THEN
                                (* DEFAULT CHARACTER IN 1 CHAR FIELD *)
                                 LDCIGEN(1) ;
                             CSPGEN(1)         (* WRC *)
                           END
                       ELSE IF LSP <> NIL
                         THEN
                           BEGIN
                             IF LSP^.FORM = SCALAR
                               THEN
                                 ERROR(399)
                             ELSE IF ISSTRING(LSP)
                               THEN
                                 BEGIN
                                   LEN := LSP^.SIZE ;
                                   IF DEFAULT
                                     THEN
                                      (* DEFAULT LENGTH OF STRING *)
                                       LDCIGEN(LEN) ;
                                   LDCIGEN(LEN) ;
                                   CSPGEN(2)    (* WRS *)
                                 END
                             ELSE
                               ERROR(116)
                           END ;
                       TEST := SY <> COMMA ;
                       IF NOT TEST
                         THEN
                           INSYMBOL
                     UNTIL TEST ;
                   END ;
                 INTEST(RPARENT,4)
              END ;
         IF LLKEY = 6
           THEN
             CSPGEN(6)                     (* WLN *)
       END ;(* WRITEP *)

       PROCEDURE NEWP ;
      (* GENERATE CODE FOR STAND PROCEDURE NEW,
          ARGUMENT OF NEW IS POINTER *)

       VAR
         LSIZE : ADDRRANGE ;

       BEGIN

         VARIABLE(FSYS +[RPARENT]) ;
        (* POINTER IS ADDRESS,PUT ON STACK *)
         LOADADDRESS ;
         LSIZE := 0 ;
         IF GATTR.TYPTR <> NIL
           THEN
             IF GATTR.TYPTR^.FORM = POINTER
               THEN
                 BEGIN
                   IF GATTR.TYPTR^.STYPE <> NIL
                     THEN
                       LSIZE := GATTR.TYPTR^.STYPE^.SIZE
                 END
               ELSE
                 ERROR(116) ;
        (* LOAD SIZE OF SPACE TO ALLOCATE IN HEAP *)
         LDCIGEN(LSIZE) ;
         CSPGEN(7)                      (* NEW *)
       END ;(* NEWP *)

       PROCEDURE RELEASEP ;
        (* GENERATE CODE TO RELEASE HEAP,
          RELEASE STARTING FROM POINTER ARGUMENT GIVEN TO RELEASE *)

       BEGIN
         VARIABLE(FSYS +[RPARENT]) ;
         IF GATTR.TYPTR <> NIL
           THEN
             IF GATTR.TYPTR^.FORM = POINTER
               THEN
                 BEGIN
                   LOAD ;
                   CSPGEN(9)               (* RST *)
                 END
               ELSE
                 ERROR(125)
       END ;(* RELEASEP *)

       PROCEDURE ORDFUNC ;
      (* GENERATE CODE FOR STANDARD FUNCTION ORD
          RESULT IS INTEGER *)

       BEGIN
         IF GATTR.TYPTR <> NIL
           THEN
             IF GATTR.TYPTR^.FORM >= POWER
               THEN
                 ERROR(125)  ;
         GATTR.TYPTR := INTPTR
       END ;(* ORDFUNC *)

       PROCEDURE CHRFUNC ;
      (* GENERATE CODE FOR STANDARD FUNCTION CHR
          RESULT IS CHARACTER *)

       BEGIN
         IF GATTR.TYPTR <> NIL
           THEN
             IF GATTR.TYPTR <> INTPTR
               THEN
                 ERROR(125) ;
         GATTR.TYPTR := CHARPTR
       END ;(* CHRFUNC *)

       PROCEDURE ODDFUNC ;
      (* GENERATE CODE FOR STANDARD FUNCTION ODD
          RESULT IS BOOLEAN *)

       BEGIN
         IF GATTR.TYPTR <> NIL
           THEN
             IF GATTR.TYPTR <> INTPTR
               THEN
                 ERROR(125) ;
         CSPGEN(12) ;                (* ODD *)
         GATTR.TYPTR := BOOLPTR
       END ;(* ODDFUNC *)

       PROCEDURE EOFFUNC ;
      (* GENERATE CODE FOR STANDARD FUNCTION EOF
          RESULT IS BOOLEAN *)

      BEGIN
        CSPGEN(8) ;            (* EOF *)
        GATTR.TYPTR := BOOLPTR ;
      END ; (* EOFFUNC *)

       PROCEDURE EOLNFUNC ;
      (* GENERATE CODE FOR STANDARD FUNCTION EOF
          RESULT IS BOOLEAN *)

      BEGIN
        CSPGEN(10) ;             (* EOF *)
	GATTR.TYPTR := BOOLPTR ;
      END ; (* EOFFUNC *)


       PROCEDURE CALLNONSTANDARD ;
      (* GENERATE TO CALL USER-DEFINED PROCEDURE OR FUNCTION,
          FCP IS POINTER TO IDENTIFIER OF PROCEDURE/FUNCTION

          STRUCTURE PARSED :

                   ------------------>--------------------
                  ^                                       !
                  !   ---       ------------       ---    V
PROC/FUNC-IDENT -----(()-----! EXPRESSION !-----())------>
                      ---   ^   ------------   !
                            !                  V
                             --------<---------                *)

       VAR
         NXT,LCP : CTP;
         LSP : STP ;
         LOCPAR,LLC : ADDRRANGE ;

       BEGIN
        (* LOCPAR IS SPACE FOR LOCAL PARAMETERS *)
         LOCPAR := 0 ;
        (* PARAMETERS LINKED AS LINKED LIST TO FCP WITH NEXT^ *)
         NXT := FCP^.NEXT ;
         PFCTTEST := FALSE ;
         IF NOT FCP^.EXTERNL
           THEN
             IF FCP^.KLASS = PROC
               THEN
                (* PROCEDURE NEEDS NO SPACE FOR RESULT *)
                 BYTEGEN(48 + LEVEL - FCP^.PFLEV)      (* MSTO *)
               ELSE
                 BEGIN
                (* FUNCTION HAS RESULT MARK STACK WITH RETURN-BYTES MSTN *)
                   BYTEGEN(64 + LEVEL - FCP^.PFLEV) ;
                   BYTEGEN(FCP^.IDTYPE^.SIZE)        (* RET VALUE *)
                 END ;
        (* PARSE PARAMETER-LIST *)
         IF SY = LPARENT
           THEN
             BEGIN
              (* REMEMBER LOCATION-COUNTER *)
               LLC := LC ;
               REPEAT
                 INSYMBOL ;
                 EXPRESSION(FSYS +[COMMA,RPARENT]) ;
                 IF GATTR.TYPTR <> NIL
                   THEN
                     BEGIN
                       IF NXT <> NIL
                         THEN
                           BEGIN
                             LSP := NXT^.IDTYPE ;
                             IF LSP <> NIL
                               THEN
                                 BEGIN
                                   IF(NXT^.VKIND = ACTUAL)
                                     THEN
                                       IF LSP^.FORM < ARRAYS
                                         THEN
                                           BEGIN
                                            (* RESERVE ROOM *)
                                             LOAD ;
                                             LOCPAR := LOCPAR + LSP^.SIZE
                                           END
                                         ELSE
                                           BEGIN
                                             IF (GATTR.KIND = EXPR) OR
                                                (GATTR.KIND = CST)
                                               THEN
                                                 BEGIN
                                                   LOAD ;
                                                   LODGEN(0,GATTR) ;
                                                   LC := LC + GATTR.TYPTR^.SIZE;
                                                   LDAGEN(0,LC) ;
                                                   IF LCMAX < LC
                                                     THEN
                                                       LCMAX := LC ;
                                                 END
                                               ELSE
                                                 LOADADDRESS ;
                                             LOCPAR := LOCPAR + PTRSIZE
					   END
	                               ELSE
                                     (* FORMAL *)
                                     IF GATTR.KIND = VARBL
                                       THEN
                                         BEGIN
                                           LOADADDRESS ;
                                           LOCPAR := LOCPAR + PTRSIZE
                                         END
                                      ELSE
                                        (* FORMAL CAN NOT BE VARIABLE *)
                                         ERROR(154) ;
                                   IF NOT COMPTYPES(LSP,GATTR.TYPTR)
                                     THEN
                                      (* ARG AND PAR NOT COMPATIBLE *)
                                       ERROR(142)
                                 END
                         END
                     END ;
                 IF NXT <> NIL
                   THEN
                    (* MORE PARAMETERS EXPECTED *)
                     NXT := NXT^.NEXT
               UNTIL SY <> COMMA ;
               LC := LLC ;
               INTEST(RPARENT,4)
             END ;(* SY = LPARENT *)
         IF NXT <> NIL
           THEN
            (* MORE PARAMETERS EXPECTED THAN FOUND *)
             ERROR(126) ;
         IF FCP^.EXTERNL
           THEN
             BEGIN
              (* EXTERNAL PROCEDURE/FUNCTION *)
               BYTEGEN(LOCPAR) ;
               BYTEGEN(187) ;               (* CAP *)
               ADDRGEN(FCP^.PFNAME)
             END
           ELSE
             GENCUP(LOCPAR,FCP^.PFNAME) ;
         IF FCP^.IDTYPE <> NIL
           THEN
             IF FCP^.IDTYPE^.SIZE = 1
               THEN
                (* RESULT FIXED TO 2 BYTES *)
                 BYTEGEN(192) ;             (* FIX21 *)
         PFCTTEST := TRUE ;
         GATTR.TYPTR := FCP^.IDTYPE
       END ;(* CALLNONSTANDARD *)

       BEGIN(* CALL *)
         IF FCP^.PFDECKIND = STANDARD
           THEN
             BEGIN
              (* STANDARD PROCEDURE AND FUNCTIONS *)
               LKEY := FCP^.KEY ;
              (* LKEY IS RETURNED BY SEARCHID *)
               IF FCP^.KLASS = PROC
                 THEN
                   (* STANDARD PROCEDURES *)
                   BEGIN
                     IF LKEY = 8
                       THEN
                        (* EXIT *)
                         BYTEGEN(161)           (* RETP *)
                     ELSE IF(LKEY = 1) OR(LKEY = 5)
                       THEN
                         READP
                     ELSE IF(LKEY = 2) OR(LKEY = 6)
                       THEN
                         WRITEP
                     ELSE
                       BEGIN
                        (* ARGUMENTS FOR PROCEDURE REQUIRED *)
                         INTEST(LPARENT,9) ;
                         IF LKEY = 3
                           THEN
                             NEWP
                         ELSE IF LKEY= 4
                           THEN
                             RELEASEP ;
                         INTEST(RPARENT,4)
                       END
                   END
                 ELSE
                  (* STANDARD FUNCTIONS *)
                   IF(LKEY < 4)
                     THEN
                       BEGIN
                        (* ARGUMENTS FOR FUNCTION REQUIRED *)
                         INTEST(LPARENT,9) ;
                         EXPRESSION(FSYS +[RPARENT]) ;
                         LOAD ;
                         CASE LKEY OF
                           1 : ORDFUNC  ;
                           2 : CHRFUNC  ;
                           3 : ODDFUNC  ;
                         END ;
                         INTEST(RPARENT,4)
                       END
                   ELSE IF(LKEY = 4)
		      THEN
	                EOLNFUNC
		      ELSE
                        EOFFUNC
             END
           ELSE
             CALLNONSTANDARD
       END ;(* CALL *)

       PROCEDURE EXPRESSION ;

         VAR
           LATTR : ATTR ;
           LOP : OPERATORN ;

         PROCEDURE SIMPLEEXPRESSION(FSYS : SETOFSYS) ;

         VAR
           LATTR : ATTR ;
           LOP : OPERATORN ;
           SIGNED : BOOLEAN ;

           PROCEDURE OPCOMP(VAR LATTR : ATTR ; OPI,OPB,OPS : INTEGER) ;
           (* PROCEDURE TO COMPILE OPERATIONS.
         	  OPI,OPB,OPS ARE THE OPERATIONS TO GENERATE
              FOR OPERANDS OF TYPE INTEGER,BOOLEAN AND SET
              RESPECTIVELY. AN ARGUMENT WILL BE ZERO FOR
              OPERAND/OPERATOR PAIRS WHICH ARE INVALID     *)

           VAR
             TEST : BOOLEAN ;

           BEGIN
             TEST := FALSE ;
             IF COMPTYPES(LATTR.TYPTR,INTPTR) AND COMPTYPES(GATTR.TYPTR,INTPTR)
               THEN
                 BEGIN
                   IF OPI <> 0
                     THEN
                       BEGIN
                         TEST := TRUE ;
                         BYTEGEN(OPI)
                       END
                 END
             ELSE IF COMPTYPES(LATTR.TYPTR,BOOLPTR) AND
                     COMPTYPES(GATTR.TYPTR,BOOLPTR)
               THEN
                 BEGIN
                   IF OPB <> 0
                     THEN
                       BEGIN
                         TEST := TRUE ;
                         BYTEGEN(OPB)
                       END
                 END
             ELSE IF(LATTR.TYPTR^.FORM = POWER) AND
                     COMPTYPES(LATTR.TYPTR,GATTR.TYPTR)
               THEN
                 IF OPS <> 0
                   THEN
                     BEGIN
                       TEST := TRUE ;
                       BYTEGEN(OPS)
                     END ;
             IF NOT TEST
               THEN
                 BEGIN
                   ERROR(134) ;
                   GATTR.TYPTR := NIL
                 END
           END ;(* OPCOMP *)

           PROCEDURE TERM(FSYS : SETOFSYS) ;

           VAR
             LATTR : ATTR ;
             LOP   : OPERATORN ;

             PROCEDURE FACTOR(FSYS : SETOFSYS) ;

             VAR
               LCP : CTP;
               LVP : CSP ;
               VARPART : BOOLEAN ;
               CSTPART : INTSET ;
               LSP : STP ;
               I   : INTEGER ;

               PROCEDURE SETEXPRESSION ;

              (* STRUCTURE PARSED :

             ---                                  ---
      --->--([)--------------------------------(])----->
             ---  !                            !  ---
                  V        ------------        ^
                   ---->--! EXPRESSION !--------
                     ^     ------------    !
                     !          ---        V
                      ---------(,)-------
                                ---                     *)

               BEGIN
                 INSYMBOL ;
                (* EMPTY SET *)
                 CSTPART :=[] ;
                 VARPART := FALSE ;
                 NEW(LSP) ;
                 LSP^.STYPE := NIL ;
                 LSP^.SIZE := SETSIZE ;
                 LSP^.FORM := POWER ;
                 IF SY = RBRACK
                   THEN
                     BEGIN
                      (* EMPTY SET *)
                       GATTR.TYPTR := LSP ;
                       GATTR.KIND := CST ;
                       INSYMBOL
                     END
                   ELSE
                     BEGIN
                       REPEAT
                         EXPRESSION(FSYS +[COMMA,RBRACK]) ;
                         IF GATTR.TYPTR <> NIL
                           THEN
                             IF GATTR.TYPTR^.FORM > SUBRANGE
                               THEN
                                 BEGIN
                                   ERROR(136) ;
                                   GATTR.TYPTR := NIL
                                 END
                               ELSE
                                 IF COMPTYPES(LSP^.STYPE,GATTR.TYPTR)
                                   THEN
                                     BEGIN
                                       IF GATTR.KIND = CST
                                         THEN
                                           CSTPART := CSTPART +
                                                      [GATTR.CVAL.IVAL]
                                         ELSE
                                           BEGIN
                                             LOAD ;
                                             BYTEGEN(174) ;   (* SGS *)
                                             IF VARPART
                                               THEN           (* UNI *)
                                                BYTEGEN(175)
                                               ELSE
                                                 VARPART := TRUE
                                            END ;
                                       LSP^.STYPE := GATTR.TYPTR ;
                                       GATTR.TYPTR := LSP ;
				     END
                                 ELSE
                                   ERROR(137) ;
                         TEST := SY <> COMMA ;
                         IF NOT TEST
                           THEN
                             INSYMBOL
                       UNTIL TEST ;
                       INTEST(RBRACK,12)
                     END ;
                 IF VARPART
                   THEN
                     BEGIN
                       IF CSTPART <>[]
                         THEN
                           BEGIN
                             LDCS(CSTPART) ;
                             BYTEGEN(175) ;        (* UNI *)
                             GATTR.KIND := EXPR
                           END
                     END
                   ELSE
                     BEGIN
                       NEW(LVP) ;
                       LVP^.PVAL := CSTPART ;
                       GATTR.CVAL.VALP := LVP
                     END
               END ;(* SETEXPRESSION *)

               BEGIN(* FACTOR *)
              (* STRUCTURE PARSED :

                ---------------------
->-------------! CONSTANT IDENTIFIER !--------------------->
   !            ---------------------                    !
   !                                                     ^
   !            ---------------------                    !
   !-----------! VARIABLE IDENTIFIER !-------------------!
   !            ---------------------                    !
   !                                                     ^
   !            ---------------------       ----------   !
   !-----------! FUNCTION IDENTIFIER !-->--! PAR-LIST !--!
   !            ---------------------       ----------   !
   !                                                     ^
   !            ---------------------                    !
   !-----------! INTEGER  CONSTANT   !-------------------!
   !            ---------------------                    !
   !                                                     ^
   !            ---------------------                    !
   !-----------! STRING CONSTANT     !-------------------!
   !            ---------------------                    !
   !                                                     ^
   !     ---    ---------------------       ---          !
   !----(()--! EXPRESSION          !-----())---------!
   !     ---    ---------------------       ---          !
   !                                                     ^
   !   -----    ---------------------                    !
   !--(NOT)--! FACTOR              !-------------------!
   !   -----    ---------------------                    !
   !                                                     ^
   !    ---     ---------------------                    !
   !---([)---! SET EXPRESSION      !-------------------
        ---     ---------------------                       *)

                 IF NOT(SY IN FACBEGSYS)
                   THEN
                     BEGIN
                       ERROR(58) ;
                       SKIP(FSYS + FACBEGSYS) ;
                       GATTR.TYPTR := NIL
                     END ;
                 WHILE SY IN FACBEGSYS DO
                   BEGIN
                     CASE SY OF
                       IDENT :
                         BEGIN
                           SEARCHID([KONST,VARS,FUNC],LCP) ;
                           INSYMBOL ;
                           IF LCP^.KLASS = FUNC
                             THEN
                               BEGIN
                                 CALL(FSYS,LCP) ;
                                 GATTR.KIND := EXPR
                               END
                           ELSE IF LCP^.KLASS = KONST
                             THEN
                               BEGIN
                                 GATTR.TYPTR := LCP^.IDTYPE ;
                                 GATTR.KIND := CST ;
                                 GATTR.CVAL := LCP^.VALUES
                               END
                           ELSE
                             SELECTOR(FSYS,LCP)
                         END ;
                       INTCONST :
                         BEGIN
                           GATTR.TYPTR := INTPTR ;
                           GATTR.KIND := CST ;
                           GATTR.CVAL := VAL ;
                           INSYMBOL
                         END ;
                       STRINGCONST :
                         BEGIN
                           IF LGTH = 1
                             THEN
                               GATTR.TYPTR := CHARPTR
                             ELSE
                               BEGIN
                                 NEW(LSP) ;
                                 LSP^.STYPE := CHARPTR ;
                                 LSP^.FORM := ARRAYS ;
                                 LSP^.INXTYPE := NIL ;
                                 LSP^.SIZE := STRINGSIZE(VAL.VALP) ;
                                 GATTR.TYPTR := LSP
                               END ;
                           GATTR.KIND := CST ;
                           GATTR.CVAL := VAL ;
                           INSYMBOL
                         END ;
                       LPARENT :
                         BEGIN
                           INSYMBOL ;
                           EXPRESSION(FSYS +[RPARENT]) ;
                           INTEST(RPARENT,4)
                         END ;
                       NOTSY :
                         BEGIN
                           INSYMBOL ;
                           FACTOR(FSYS) ;
                           LOAD ;
                           BYTEGEN(172) ;     (* NOT *)
                           IF GATTR.TYPTR <> NIL
                             THEN
                               IF NOT COMPTYPES(GATTR.TYPTR,BOOLPTR)
                                 THEN
                                   BEGIN
                                    (* RESULT MUST BE
                                        BOOLEAN TO BE NEGATED *)
                                     ERROR(135) ;
                                     GATTR.TYPTR := NIL
                                   END ;
                         END ;
                       LBRACK : SETEXPRESSION
                     END ;
                     TEST2(FSYS,6,FACBEGSYS)
                   END ;(* WHILE *)
                 END ;(* FACTOR *)

           BEGIN(* TERM *)
          (* STRUCTURE PARSED :

         --------
 ---->--! FACTOR !-------------------------------------------------->
         --------  ^               !     !      !       !       !
                   !               V     V      V       V       V
                   !              ---   ---   -----   -----   -----
                   !             PLUS(/)(DIV)(MOD)(AND)
                   !              ---   ---   -----   -----   -----
                   !   --------    !     !      !       !       !
                    <-! FACTOR !-<----<-----<-------<------<----
                       --------                                     *)

             FACTOR(FSYS +[MULOP]) ;
             WHILE SY = MULOP DO
               BEGIN
                 LOAD ;
                 LATTR := GATTR ;
                 LOP := OP ;
                 INSYMBOL ;
                 FACTOR(FSYS +[MULOP]) ;
                 LOAD ;
                 IF(LATTR.TYPTR <> NIL) AND (GATTR.TYPTR <> NIL)
                   THEN
                     BEGIN
                       IF LOP = MUL
                         THEN
                           OPCOMP(LATTR,170,0,167)(* MPI,INT *)
                       ELSE IF LOP = IDIV
                         THEN
                           OPCOMP(LATTR,165,0,0  )(* DVI *)
                       ELSE IF LOP = IMOD
                         THEN
                           OPCOMP(LATTR,169,0,0  )(* MOD *)
                       ELSE IF LOP = ANDOP
                         THEN
                           OPCOMP(LATTR,0,163, 0  )(* AND *)
                     END
                   ELSE
                     GATTR.TYPTR := NIL
               END ;(* WHILE *)
           END ;(* TERM *)

         BEGIN(* SIMPLEEXPRESSION *)
        (* STRUCTURE PARSED :

            ---
        ---(+)->-
       !    ---    !
       !           V     ------
---->--!------------->--! TERM !---------------------------------->
       !           ^     ------  ^              !      !        !
       !    ---    !             !              V      V        V
        ---(-)->-              !             ---    ---      ----
            ---                  !           (+) (-)   (OR)
                                 !             ---    ---      ----
                                 !   ------     !      !        !
                                  --! TERM !--------------------
                                     ------                       *)

           SIGNED := FALSE ;
           IF(SY = ADDOP) AND (OP IN[PLUS,MINUS])
             THEN
               BEGIN
                 SIGNED := (OP = MINUS) ;
                 INSYMBOL
               END ;
           TERM(FSYS +[ADDOP]) ;
           IF SIGNED
             THEN
               BEGIN
                 LOAD ;
                 IF COMPTYPES(GATTR.TYPTR,INTPTR)
                   THEN
                     BYTEGEN(171)                (* NGI *)
                   ELSE
                     BEGIN
                       ERROR(134) ;
                       GATTR.TYPTR := NIL
                     END
               END ;
           WHILE SY = ADDOP DO
             BEGIN
              (* SAVE ON STACK *)
               LOAD ;
              (* REMEMBER PREVIOUS EXPRESSION RESULT *)
               LATTR := GATTR ;
               LOP := OP ;
               INSYMBOL ;
               TERM(FSYS +[ADDOP]) ;
               LOAD ;
               IF(LATTR.TYPTR <> NIL) AND (GATTR.TYPTR <> NIL)
                 THEN
                   BEGIN
                     IF LOP = PLUS
                       THEN
                         OPCOMP(LATTR,162,0,175) (* ADI,UNI *)
                     ELSE IF LOP = MINUS
                       THEN
                         OPCOMP(LATTR,173,0,164) (* SBI,DIF *)
                     ELSE IF LOP = OROP
                       THEN
                         OPCOMP(LATTR,0,168,0 ) (* OR *)
                   END
                 ELSE
                   GATTR.TYPTR := NIL ;
               GATTR.KIND := EXPR
             END ;(* WHILE SY = ADDOP *)
         END ;(* SIMPLEEXPRESSION *)

       BEGIN(* EXPRESSION *)
      (* STRUCTURE PARSED :
  -------------
-! SIMPLE EXPR !------------------------------------------------------->
  -------------  !                                                    !
                 V                                                    !
               --------------------------------------                 !
              !     !     !     !      !      !      !                !
              V     V     V     V      V      V      V                !
             ---   ---   ---   ----   ----   ----   ----              !
           (=)(<)(>)(<>)(>=)(<=)(IN)             ^
             ---   ---   ---   ----   ----   ----   ----              !
              !     !     !     !      !      !      !                !
              V     V     V     V      V      V      V  ------------  !
               ----------------------------------------! SIMPL EXPR !-
                                                        ------------  *)

         SIMPLEEXPRESSION(FSYS +[RELOP]) ;
         IF SY = RELOP
           THEN
             BEGIN
               IF GATTR.TYPTR <> NIL
                 THEN
                   IF GATTR.TYPTR^.FORM <= POWER
                     THEN
                       LOAD
                     ELSE
                       LOADADDRESS ;
               LATTR := GATTR ;
               LOP := OP ;
               INSYMBOL ;
               SIMPLEEXPRESSION(FSYS) ;
               IF GATTR.TYPTR <> NIL
                 THEN
                   IF GATTR.TYPTR^.FORM <= POWER
                     THEN
                       LOAD
                     ELSE
                       LOADADDRESS ;
               IF(LATTR.TYPTR <> NIL) AND(GATTR.TYPTR <> NIL)
                 THEN
                   IF LOP = INOP
                     THEN
					   BEGIN
                         IF GATTR.TYPTR^.FORM = POWER
                           THEN
                             IF COMPTYPES(LATTR.TYPTR, GATTR.TYPTR^.STYPE)
                               THEN
                                 BYTEGEN(166)             (* INN *)
                               ELSE
                                 BEGIN
                                   ERROR(129) ;
                                   GATTR.TYPTR := NIL
                                 END
                           ELSE
                             BEGIN
                               ERROR(130) ;
                               GATTR.TYPTR := NIL
                             END
					   END
                   ELSE
                       BEGIN
                         IF COMPTYPES(LATTR.TYPTR,GATTR.TYPTR)
                           THEN
                             BEGIN
                               IF LATTR.TYPTR^.FORM = POINTER
                                 THEN
                                   BEGIN
                                    (* TEST ON EQUALITY ALLOWED ONLY *)
                                     IF LOP IN[LTOP,LEOP,GTOP,GEOP]
                                       THEN
                                         ERROR(131)
                                   END
                               ELSE IF LATTR.TYPTR^.FORM = POWER
                                 THEN
                                   BEGIN
                                    (* INCLUSION NOT ALLOWED IN SET COMP *)
                                     IF LOP IN[LTOP,GTOP]
                                       THEN
                                         ERROR(132)
                                   END
                               ELSE IF LATTR.TYPTR^.FORM = ARRAYS
                                 THEN
                                   BEGIN
                                    (* TEST ON EQUALITY ALLOWED ONLY *)
                                     IF NOT ISSTRING(LATTR.TYPTR) AND
                                       (LOP IN[LTOP,LEOP,GTOP,GEOP])
                                       THEN
                                         ERROR(131)
                                   END
                               ELSE IF LATTR.TYPTR^.FORM = RECORDS
                                 THEN
                                   BEGIN
                                    (* TEST ON EQUALITY ALLOWED ONLY *)
                                     IF LOP IN[LTOP,LEOP,GTOP,GEOP]
                                       THEN
                                         ERROR(131)
                                   END ;
                            IF LOP = LTOP
                                 THEN
                                   CONDGEN(147,LATTR) (* LES *)
                            ELSE IF LOP = LEOP
                               THEN
                                 CONDGEN(144,LATTR) (* LEQ *)
                            ELSE IF LOP = EQOP
                              THEN
                                CONDGEN(150,LATTR) (* EQU *)
                            ELSE
                               BEGIN
                                 IF LOP = GTOP
                                   THEN
                                     CONDGEN(144,LATTR)(* LEQ *)
                                 ELSE IF LOP = GEOP
                                   THEN
                                     CONDGEN(147,LATTR)(* LES *)
                                 ELSE IF LOP = NEOP
                                   THEN
                                     CONDGEN(150,LATTR)(* EQU *) ;
                                 BYTEGEN(172)        (* NOT *)
                               END
                             END
                           ELSE
                             ERROR(129)
                       END ;(* LOP = INOP *)
                   GATTR.TYPTR := BOOLPTR ;
                   GATTR.KIND := EXPR
             END (* SY = RELOP *)
        END ;(* EXPRESSION *)

        PROCEDURE ASSIGNMENT(FCP : CTP) ;
       (* STRUCTURE PARSED :

                ------------       ---      ------------
        ->-----! IDENTIFIER !-----(=)----! EXPRESSION !---->
                ------------       ---      ------------
                                                           *)

        VAR
          LATTR : ATTR ;

        BEGIN
         (* DETERMINE TYPE OF IDENTIFIER FOR ASSIGNMENT *)
          SELECTOR(FSYS +[BECOMES],FCP) ;
          IF SY = BECOMES
            THEN
              BEGIN
                IF GATTR.TYPTR <> NIL
                  THEN
                    IF(GATTR.ACCESS <> DRCT) OR (GATTR.TYPTR^.FORM > POWER)
                      THEN
                        LOADADDRESS ;
                LATTR := GATTR ;
                INSYMBOL ;
                EXPRESSION(FSYS) ;
                IF GATTR.TYPTR <> NIL
                  THEN
                   (* COMPLEX STRUCTURE NOT SIMPLE STORE *)
                    IF GATTR.TYPTR^.FORM  <= POWER
                      THEN
                        LOAD
                      ELSE
                        LOADADDRESS ;
                IF(LATTR.TYPTR <> NIL) AND (GATTR.TYPTR <> NIL)
                  THEN
                    BEGIN
                      IF COMPTYPES(LATTR.TYPTR,GATTR.TYPTR)
                        THEN
                          BEGIN
                            IF LATTR.TYPTR^.FORM IN [ARRAYS,RECORDS]
                              THEN
                                BEGIN
                                  BYTEGEN(183) ;           (* MOV *)
                                  ADDRGEN(LATTR.TYPTR^.SIZE)
                                END
                              ELSE
                                STORE(LATTR)
                          END
                        ELSE
                          ERROR(129)
                    END
              END(* SY = BECOMES *)
            ELSE
             (* := EXPECTED *)
              ERROR(51)
        END ;(* ASSIGNMENT *)

        PROCEDURE COMPOUNDSTATEMENT ;
       (* STRUCTURE PARSED :

             -------       -----------        -----
   ---->---(BEGIN)-----! STATEMENT !------(END)---->
             -------   ^   -----------   !    -----
                       !                 V
                       !       ---       !
                        ------(;)------
                               ---                       *)
        BEGIN
          REPEAT
            REPEAT
              STATEMENT(FSYS +[SEMICOLON,ENDSY])
            UNTIL NOT(SY IN STATBEGSYS) ;
            TEST := SY <> SEMICOLON ;
            IF NOT TEST
              THEN
                INSYMBOL
          UNTIL TEST ;
          INTEST(ENDSY,13)
        END ;(* COMPOUNDSTATEMENT *)

        PROCEDURE IFSTATEMENT ;
       (* PARSED BY IFSTATEMENT :
          (IF SYMBOL ALREADY PARSED)
                                               ----   ---------
                                            --(ELSE)-!STATEMENT!-
                                           !   ----   ---------  !
         ----------    ----    ---------   !                     V
 IF ->--!EXPRESSION!--(THEN)--!STATEMENT!--!                     -->
         ----------    ----    ---------   !                     ^
                                           !                     !
                                            ---------------------

                                                *)

        VAR
          ICIX1,ICIX2 : INTEGER ;

        BEGIN
         (* EXPRESSION MUST GIVE BOOLEAN RESULT FOR COND. JUMP *)
          EXPRESSION(FSYS +[THENSY]) ;
          GENFJP(0) ;
         (* REMEMBER ADDRESS FOR JUMP TO PLANT AFTER STATEMENT *)
          ICIX1 := IC + ICN - 2 ;
         (* THEN SYMBOL *)
          INTEST(THENSY,52) ;
         (* STATEMENT *)
         STATEMENT(FSYS +[ELSESY]) ;
         (* MAYBE ELSE PART WITH STATEMENT *)
          IF SY = ELSESY
            THEN
              BEGIN
               (* END THEN PART WITH UNCONDITIONAL JUMP TO END *)
                ICIX2 := IC + ICN ;
                GENUJPENT(178,0) ;    (* UJP *)
               (* FILL IN ADDRESS OF COND JUMP IF RESULT IS FALSE *)
                PLANTADDR( ICIX1,IC + ICN) ;
                INSYMBOL ;
                STATEMENT(FSYS) ;
               (* WE KNOW NOW END OF ELSE PART *)
                PLANTADDR(ICIX2 + 1,IC + ICN)
              END
            ELSE
             (* NO ELSE PART,FILL IN CONDITIONAL JUMP ADDRESS *)
              PLANTADDR(ICIX1,IC + ICN)
        END ;(* IFSTATEMENT *)

        PROCEDURE CASESTATEMENT ;
       (* PARSED BY CASESTATEMENT :
          (CASE SYMBOL ALREADY PARSED)
            -----------     ----
CASE->----! EXPRESSION !---(OF) ---->---
           ------------     ----          !
                                          !
    --------------------------------------
   !
   V
   !
    -----------! CONSTANT !------(;) ---! STATEMENT !---------
     ^     ^    ----------   !    ---      -----------       !  !
     !     !                 V                               V  V
     !     !        ---      !                               !  !
     !      -------(,)-----                                !  !
     !              ---              ---                     !  !
      ------------------------------(;)--------------------   !
                                     ---                        !
    ------------------------------------------------------------
   !
   V                ------
   !          -----(ELSE)----
   !         !      ------     !
   !         !                 !     -----------
    ---->----!                 !-->-! STATEMENT ! -->-------->
   !         !                 !     -----------         !
   !                                                     !
    ---->------------------------------------------------  *)

        TYPE
          CIP = ^CASEINFO ;
          CASEINFO    = PACKED RECORD
                          NEXT : CIP ;
                          CSSTART,CSLAB : INTEGER
                        END ;

        VAR
          LSP,LSP1 : STP ;
          FSTPTR,LPT1,LPT2,LPT3 : CIP ;
          LVAL : VALU ;
          LCIX,LCIX1,LMIN,LMAX : INTEGER ;

        BEGIN
          EXPRESSION(FSYS +[OFSY,COMMA,COLON]) ;
          LOAD ;
          LCIX := IC + ICN ;
          GENUJPENT(178,0) ;            (* UJP *)
          GENUJPENT(178,0) ;            (* UJP *)
          LSP := GATTR.TYPTR ;
          IF LSP <> NIL
            THEN
              IF LSP^.FORM > SUBRANGE
                THEN
                  BEGIN
                    ERROR(144) ;
                    LSP := NIL
                  END ;
          INTEST(OFSY,8) ;
          FSTPTR := NIL ;
          REPEAT
            LPT3 := NIL ;
            LCIX1 := IC + ICN ;
            REPEAT
              CONSTANT(FSYS +[COMMA,COLON],LSP1,LVAL) ;
              IF LSP1 <> NIL
                THEN
                  IF COMPTYPES(LSP,LSP1)
                    THEN
                      BEGIN
                        LPT1 := FSTPTR ;
                        LPT2 := NIL ;
                        TEST := TRUE ;
                        WHILE(LPT1 <> NIL) AND TEST DO
                          IF LPT1^.CSLAB <= LVAL.IVAL
                            THEN
                              BEGIN
                                IF LPT1^.CSLAB = LVAL.IVAL
                                  THEN
                                    ERROR(156) ;
                                TEST := FALSE
                              END
                            ELSE
                              BEGIN
                                LPT2 := LPT1 ;
                                LPT1 := LPT1^.NEXT
                              END ;
                        NEW(LPT3) ;
                        LPT3^.NEXT := LPT1 ;
                        LPT3^.CSLAB := LVAL.IVAL ;
                        LPT3^.CSSTART := LCIX1 ;
                        IF LPT2 = NIL
                          THEN
                            FSTPTR := LPT3
                          ELSE
                            LPT2^.NEXT := LPT3
                      END
                    ELSE
                      ERROR(147) ;
                TEST := SY <> COMMA ;
                IF NOT TEST
                  THEN
                    INSYMBOL
            UNTIL TEST ;
            INTEST(COLON,5) ;
            REPEAT
              STATEMENT(FSYS +[SEMICOLON,ELSESY])
            UNTIL NOT(SY IN STATBEGSYS) ;
            IF LPT3 <> NIL
              THEN
                GENUJPENT(178,LCIX + 3) ;   (* UJP *)
            TEST := SY <> SEMICOLON ;
            IF NOT TEST
              THEN
                INSYMBOL ;
          UNTIL TEST OR(SY = ENDSY) OR(SY = ELSESY) ;
          PLANTADDR(LCIX + 1,IC + ICN) ;
          IF FSTPTR <> NIL
            THEN
              BEGIN
                LMAX := FSTPTR^.CSLAB ;
                LPT1 := FSTPTR ;
                FSTPTR := NIL ;
               (* REVERSE POINTERS *)
                REPEAT
                  LPT2 := LPT1^.NEXT ;
                  LPT1^.NEXT := FSTPTR ;
                  FSTPTR := LPT1 ;
                  LPT1 := LPT2
                UNTIL LPT1 = NIL ;
                LMIN := FSTPTR^.CSLAB ;
                BYTEGEN(182)  ;         (* CAS *)
                ADDRGEN(LMIN) ;
                ADDRGEN(LMAX) ;
                LCIX1 := IC + ICN ;
                ADDRGEN(0) ;
                REPEAT
                  WHILE FSTPTR ^.CSLAB > LMIN DO
                    BEGIN
                      ADDRGEN(0) ;
                      LMIN := LMIN + 1 ;
                    END ;
                  ADDRGEN(FSTPTR^.CSSTART) ;
                  FSTPTR := FSTPTR^.NEXT ;
                  LMIN := LMIN + 1
                UNTIL FSTPTR = NIL
              END(* FSTPTR <> NIL *)
            ELSE
              ERROR(157) ;
          IF SY = ELSESY
            THEN
              BEGIN
                INSYMBOL ;
                PLANTADDR(LCIX1,IC + ICN) ;
                REPEAT
                  STATEMENT(FSYS +[SEMICOLON])
                UNTIL NOT(SY IN STATBEGSYS) ;
                GENUJPENT(178,LCIX + 3)
              END ;
          PLANTADDR(LCIX + 4,IC + ICN) ;
          INTEST(ENDSY,13)
        END ;(* CASESTATEMENT *)

        PROCEDURE REPEATSTATEMENT ;
       (* PARSED BY REPEATSTATEMENT :
          (REPEAT SYMBOL ALREADY PARSED)

             -----------         -------       ------------
REPEAT->----! STATEMENT !-------(UNTIL) ----! EXPRESSION !--->
         ^   -----------     !   -------       ------------
         !         -         !
          --------(;)--------
                   -
                                                *)
          VAR
            LADDR : INTEGER ;

          BEGIN
           (* REMEMBER ADDRESS FOR FALSE JUMP AT UNTIL *)
            LADDR := IC + ICN ;
           (* ONE OR MORE STATEMENT BLOCKS *)
            REPEAT
              REPEAT
                STATEMENT(FSYS +[SEMICOLON,UNTILSY])
              UNTIL NOT(SY IN STATBEGSYS) ;
              TEST := SY <> SEMICOLON ;
              IF NOT TEST
                THEN
                  INSYMBOL
            UNTIL TEST ;
           (* UNTIL PART *)
            IF SY = UNTILSY
              THEN
                BEGIN
                  INSYMBOL ;
                 (* EXPRESSION *)
                  EXPRESSION(FSYS) ;
                 (* EXPRESSION RESULT MUST BE BOOLEAN FOR FJP *)
                  GENFJP(LADDR)
                END
              ELSE
               (* UNTIL EXPECTED BUT NOT FOUND *)
                ERROR(53)
        END ;(* REPEATSTATEMENT *)

        PROCEDURE WHILESTATEMENT ;
       (* PARSED BY WHILESTATEMENT :
          (WHILE SYMBOL ALREADY PARSED)

              ------------           ----       ----------
 WHILE ->----! EXPRESSION !---------(DO) ----!STATEMENT !--->
              ------------           ----       ----------

                                                *)
        VAR
          LADDR,LCIX : INTEGER ;

        BEGIN
         (* REMEMBER ADDRESS FOR JUMP AT AT END OF WHILE *)
          LADDR := IC + ICN ;
         (* PARSE BOOLEAN EXPRESSION *)
          EXPRESSION(FSYS +[DOSY]) ;
         (* REMEMBER ADDRESS OF CONDITIONAL JUMP OUT OF LOOP *)
          LCIX := IC + ICN ;
         GENFJP(0) ;
         (* DO SYMBOL EXPECTED *)
          INTEST(DOSY,54) ;
         (* PARSE BODY OF LOOP *)
          STATEMENT(FSYS) ;
         (* UNCONDITIONAL JUMP BACK TO BEGIN OF LOOP *)
          GENUJPENT(178,LADDR) ;    (* UJP *)
         (* ADDRESS FOR JUMP OUT OF LOOP NOW KNOWN *)
          PLANTADDR(LCIX + 1,IC + ICN) ;
        END ;(* WHILESTATEMENT *)

        PROCEDURE FORSTATEMENT ;
       (* PARSED BY FORSTATEMENT :
          (FOR SYMBOL ALREADY PARSED)

         -----------    ----     ------------
FOR ->--! VAR-IDENT !--(:=)---! EXPRESSION !--->
         -----------    ----     ------------     !
                                                  !
  ------------------------------------------------
 !        -------
 !    --(DOWNTO)---
 !   !   --------    !   ------------    ----    -----------
  ->-!               !--! EXPRESSION !--(DO)--! STATEMENT !--->
     !     ----      !   ------------    ----    -----------
      ---(TO)-----
           ----                                 *)

        VAR
          LATTR : ATTR ;
          LSP : STP ;
          LSY : SYMBOLN ;
          LCIX,LADDR : INTEGER ;

        BEGIN
          (* VARIABLE IDENTIFIER,FIND DESCRIPTION *)
          IF SY = IDENT
            THEN
              BEGIN
                SEARCHID([VARS],LCP) ;
                LATTR.TYPTR := LCP^.IDTYPE ;
                LATTR.KIND := VARBL ;
                IF LCP^.VKIND = ACTUAL
                  THEN
                    LDAGEN(LEVEL - LCP^.VLEV,LCP^.VADDR)
                  ELSE
                    BEGIN
                      ERROR(155) ;
                      LATTR.TYPTR := NIL
                    END ;
                IF LATTR. TYPTR <> NIL
                  THEN
                    IF(LATTR.TYPTR^.FORM > SUBRANGE)
                      THEN
                        BEGIN
                         (* ILLEGAL TYPE OF LOOP CONTROL VARIABLE *)
                          ERROR(143) ;
                          LATTR.TYPTR := NIL
                        END ;
                INSYMBOL ;
              END
            ELSE
              BEGIN
               (* IDENTIFIER EXPECTED *)
                ERROR(2) ;
                SKIP(FSYS +[BECOMES,TOSY,DOWNTOSY,DOSY])
              END ;
         (* := *)
          IF SY = BECOMES
            THEN
              BEGIN
                INSYMBOL ;
               (* EXPRESSION,MUST BE SAME TYPE AS VAR. IDENT *)
                EXPRESSION(FSYS +[TOSY,DOWNTOSY,DOSY]) ;
                IF GATTR.TYPTR <> NIL
                  THEN
                    IF GATTR.TYPTR^.FORM > SUBRANGE
                      THEN
                       (* ILLEGAL TYPE OF EXPRESSION *)
                        ERROR(144)
                      ELSE
                        IF COMPTYPES(LATTR.TYPTR,GATTR.TYPTR)
                          THEN
                           (* PUT BEGINVALUE OF LOOP VARIABLE ON STACK *)
                            LOAD
                          ELSE
                           (* TYPE CONFLICT *)
                            ERROR(145)
              END
            ELSE
              BEGIN
               (* := EXPECTED *)
                ERROR(51) ;
                SKIP(FSYS +[TOSY,DOWNTOSY,DOSY])
              END ;
         (* PARSE TO OR DOWNTO *)
          IF SY IN[TOSY,DOWNTOSY]
            THEN
              BEGIN
                LSY := SY ;
                INSYMBOL ;
                EXPRESSION(FSYS +[DOSY]) ;
                IF GATTR.TYPTR <> NIL
                  THEN
                    IF GATTR.TYPTR^.FORM > SUBRANGE
                      THEN
                        ERROR(144)
                      ELSE
                        IF COMPTYPES(LATTR.TYPTR,GATTR.TYPTR)
                          THEN
                            BEGIN
                             (* PUT END VALUE OF LOOP VARIABLE ON STACK *)
                              LOAD ;
                              LCIX := LATTR.TYPTR^.SIZE - 1 ;
                              IF LSY = DOWNTOSY
                                THEN
                                 LCIX := LCIX + 4 ;
                             (* STEPSIZE ON STACK *)
                              BYTEGEN(LCIX) ;
                              LADDR := IC + ICN ;
                             (* GENERATE JUMP INSTRUCTION TO FOR STATEMENT *)
                              GENUJPENT(178,0)     (* UJP *)
                            END
                          ELSE
                           (* TYPE CONFLICT WITH LOOP VARIABLE *)
                            ERROR(145)
              END
            ELSE
              BEGIN
               (* TO/DOWNTO EXPECTED *)
                ERROR(55) ;
                SKIP(FSYS +[DOSY])
              END ;
         (* SKIP AND TEST DO SYMBOL *)
          INTEST(DOSY,54) ;
         (* STATEMENT *)
          STATEMENT(FSYS) ;
         (* NOW WE KNOW ADDRESS OF END OF FOR LOOP *)
          PLANTADDR(LADDR + 1,IC + ICN) ;
          GENUJPENT(145,LADDR + 3)  ;            (* FOR *)
        END ;(* FORSTATEMENT *)

      BEGIN(* STATEMENT *)
        IF NOT(SY IN FSYS +[IDENT])
          THEN
            BEGIN
              ERROR(6) ;
              SKIP(FSYS)
            END ;
        IF SY  IN(STATBEGSYS +[IDENT])
          THEN
            BEGIN
              PFCTTEST := FALSE ;
              CASE SY OF
                IDENT :
                  BEGIN
                    SEARCHID([VARS,FUNC,PROC],LCP) ;
                    INSYMBOL ;
                    IF LCP^.KLASS = PROC
                      THEN
                        CALL(FSYS,LCP)
                      ELSE
                        ASSIGNMENT(LCP)
                  END ;
                BEGINSY :
                  BEGIN
                    INSYMBOL ;
                    COMPOUNDSTATEMENT
                  END ;
                IFSY :
                  BEGIN
                    INSYMBOL ;
                    IFSTATEMENT
                  END ;
                WHILESY :
                  BEGIN
                    INSYMBOL ;
                    WHILESTATEMENT
                  END ;
                REPEATSY :
                  BEGIN
                    INSYMBOL ;
                    REPEATSTATEMENT
                  END ;
                FORSY :
                  BEGIN
                    INSYMBOL ;
                    FORSTATEMENT
                  END ;
                CASESY :
                  BEGIN
                    INSYMBOL ;
                    CASESTATEMENT
                  END
              END ;
              IF NOT(SY IN[SEMICOLON,ENDSY,ELSESY,UNTILSY])
                THEN
                  BEGIN
                    ERROR(6) ;
                    SKIP(FSYS)
                  END
            END ;
      END ;(* STATEMENT *)

    BEGIN(* BODY *)
     (* FPROCP IS POINTER TO PROC/FUNC/PROGRAM IDENTIFIER *)
      ENTNAME := FPROCP^.PFNAME ;
     (* EMPTY CODE BUFFER *)
      WRITEOUT ;
      SUMCHECK := 0 ;
      WRITE('P4') ;
      HEXOUT(ENTNAME) ;
      FOR I := 1 TO 8 DO
        BEGIN
          WRITE(FPROCP^.NAME[I]) ;
          SUMCHECK := SUMCHECK + ORD(FPROCP^.NAME[I])
        END ;
      HEXOUT((16383 - SUMCHECK) MOD 256) ;
      WRITELN ;
     (* REMEMBER WHERE TO PUT NR OF BYTES FOR LOCAL VARIABLES *)
      SEGSIZE := IC + ICN ;
      GENUJPENT(181,0) ;              (* ENT *)
     (* START WITH 6 BYTES ON STACK *)
      LLC1 := LCAFTERMARKSTACK ;
     (* PARAMETER-LIST AS LINKED LIST ON NEXT^ *)
      LCP := FPROCP^.NEXT ;
      WHILE LCP <> NIL DO
        BEGIN
          IF LCP^.KLASS = VARS
            THEN
              IF LCP^.IDTYPE <> NIL
                THEN
                  IF (LCP^.VKIND = ACTUAL) AND (LCP^.IDTYPE^.FORM > POWER)
                    THEN
                      BEGIN
                        LLC1 := LLC1 + PTRSIZE ;
                        (* COPY RECORD OR ARRAY *)
                        LDAGEN(0,LCP^.VADDR) ;
                        BYTEGEN(96) ;(* LOD *)
                        BYTEGEN(LLC1) ;
                        BYTEGEN(183) ;    (* MOV *)
                        ADDRGEN(LCP^.IDTYPE^.SIZE)
                      END
                    ELSE
                     (* KEEP TRACK OF LOCAL SPACE ON STACK *)
                      LLC1 := LLC1 + LCP^.IDTYPE^.SIZE ;
              LCP:= LCP^.NEXT ;
        END ;(* WHILE *)
      LCMAX := LC ;
      REPEAT
        REPEAT
          STATEMENT(FSYS +[SEMICOLON,ENDSY])
        UNTIL NOT(SY IN STATBEGSYS) ;
        TEST := SY <> SEMICOLON ;
        IF NOT TEST
          THEN
            INSYMBOL
      UNTIL TEST ;
      INTEST(ENDSY,13) ;
      BYTEGEN(161) ;               (* RETP *)
      PLANTADDR(SEGSIZE + 1,LCMAX - LCAFTERMARKSTACK) ;
    END ;(* BODY *)

  BEGIN(* BLOCK*)

    DP := TRUE ;
    REPEAT
      IF SY = CONSTSY
        THEN
          BEGIN
            INSYMBOL ;
            CONSTDECLARATION
          END  ;
      IF SY = TYPESY
        THEN
          BEGIN
            INSYMBOL ;
            TYPEDECLARATION
          END  ;
      IF SY = VARSY
        THEN
          BEGIN
            INSYMBOL ;
            VARDECLARATION
          END  ;
      WHILE SY IN[PROCSY,FUNCSY] DO
        BEGIN
          LSY := SY ;
          INSYMBOL ;
          PROCDECLARATION(LSY)
        END ;
      IF SY <> BEGINSY
        THEN
          BEGIN
            ERROR(18) ;
            SKIP(FSYS)
          END
    UNTIL SY IN STATBEGSYS ;

    DP := FALSE ;
    INTEST(BEGINSY,17) ;
    REPEAT
      BODY(FSYS) ;
      TEST2([FSY],6,FSYS)
    UNTIL(SY = FSY) OR(SY IN BLOCKBEGSYS) ;
  END ;(* BLOCK*)

  PROCEDURE STDNAMES ;

  BEGIN
    NA[0]  := 'FALSE   ' ; NA[1]  := 'TRUE    ' ; NA[2]  := 'READ    ' ;
    NA[3]  := 'WRITE   ' ;
    NA[4]  := 'NEW     ' ; NA[5]  := 'RELEASE ' ; NA[6]  := 'READLN  ' ;
    NA[7]  := 'WRITELN ' ; NA[8]  := 'ORD     ' ; NA[9]  := 'CHR     ' ;
    NA[10] := 'ODD     ' ; NA[11] := 'EOLN    ' ; NA[12] := 'EOF     ' ;
  END ;(* STDNAMES *)

  PROCEDURE ENTERSTDTYPES ;

  BEGIN
   (* INTEGER *)
    NEW(INTPTR) ;
    INTPTR^.SIZE  := INTSIZE ;
    INTPTR^.FORM  := SCALAR  ;
    INTPTR^.STYPE := NIL     ;
   (* CHAR *)
    NEW(CHARPTR) ;
    CHARPTR^.SIZE  := CHARSIZE ;
    CHARPTR^.FORM  := SCALAR  ;
    CHARPTR^.STYPE := NIL     ;
   (* BOOLEAN *)
    NEW(BOOLPTR) ;
    BOOLPTR^.SIZE  := BOOLSIZE ;
    BOOLPTR^.FORM  := SCALAR  ;
    BOOLPTR^.STYPE := NIL     ;
   (* POINTER NIL *)
    NEW(NILPTR) ;
    NILPTR^.SIZE  := PTRSIZE ;
    NILPTR^.FORM  := POINTER ;
    NILPTR^.STYPE := NIL     ;
  END ;(* ENTERSTDTYPES *)

  PROCEDURE ENTERSTNAMES ;

  VAR
      CP,CP1 : CTP;
      I : INTEGER ;

  BEGIN
   (* INTEGER *)
    NEW(CP) ;
    CP^.NAME := 'INTEGER ' ;
    CP^.IDTYPE := INTPTR ;
    CP^.NEXT := NIL ;
    CP^.KLASS := TYPES ;
    ENTERID(CP) ;
   (* CHAR *)
    NEW(CP) ;
    CP^.NAME := 'CHAR    ' ;
    CP^.IDTYPE := CHARPTR ;
    CP^.NEXT := NIL ;
    CP^.KLASS := TYPES ;
    ENTERID(CP) ;
   (* BOOLEAN *)
    NEW(CP) ;
    CP^.NAME := 'BOOLEAN ' ;
    CP^.IDTYPE := BOOLPTR ;
    CP^.NEXT := NIL ;
    CP^.KLASS := TYPES ;
    ENTERID(CP) ;
   (* FALSE,TRUE *)
    CP1 := NIL ;
    FOR I := 0 TO 1 DO
      BEGIN
        NEW(CP) ;
        CP^.NAME := NA[I] ;
        CP^.IDTYPE := BOOLPTR ;
        CP^.NEXT := CP1 ;
        CP^.KLASS := KONST ;
          CP^.VALUES.IVAL := I ;
        ENTERID(CP) ;
        CP1 := CP
      END ;
   (* NIL POINTER *)
    NEW(CP) ;
    CP^.NAME := 'NIL     ' ;
    CP^.IDTYPE := NILPTR ;
    CP^.NEXT := NIL ;
    CP^.KLASS := KONST ;
      CP^.VALUES.IVAL := 0 ;
    ENTERID(CP) ;
   (* I/O AND POINTER PROCEDURES
       KEY  1 = READ
       KEY  2 = WRITE
       KEY  3 = NEW
       KEY  4 = RELEASE
       KEY  5 = READLN
       KEY  6 = WRITELN
   *)
    FOR I := 2 TO 7 DO
      BEGIN
        NEW(CP) ;
        CP^.NAME := NA[I] ;
        CP^.IDTYPE := NIL ;
        CP^.NEXT := NIL ;
        CP^.KLASS := PROC ;
        CP^.PFDECKIND := STANDARD ;
        CP^.KEY := I - 1 ;
        ENTERID(CP)
      END ;
   (* STANDARD FUNCTIONS :
       KEY 1 = ORD
       KEY 2 = CHR
       KEY 3 = ODD
       KEY 6 = EOLN
       KEY 7 = EOF
	*)
    FOR I := 8 TO 12 DO
      BEGIN
        NEW(CP) ;
        CP^.NAME := NA[I] ;
        CP^.IDTYPE := NIL ;
        CP^.NEXT := NIL ;
        CP^.KLASS := FUNC ;
        CP^.PFDECKIND := STANDARD ;
        CP^.KEY := I - 7 ;
        ENTERID(CP)
      END ;
  END ;(* ENTERSTNAMES *)

  PROCEDURE ENTERUNDECL ;

  BEGIN
    NEW(UTYPPTR) ;
    UTYPPTR^.NAME := '        ' ;
    UTYPPTR^.IDTYPE := NIL ;
    UTYPPTR^.KLASS := TYPES ;
    NEW(UCSPTR) ;
    UCSPTR^.NAME := '        ' ;
    UCSPTR^.IDTYPE := NIL ;
    UCSPTR^.NEXT := NIL ;
    UCSPTR^.KLASS := KONST ;
      UCSPTR^.VALUES.IVAL := 0 ;
    NEW(UVARPTR) ;
    UVARPTR^.NAME := '        ' ;
    UVARPTR^.IDTYPE := NIL ;
    UVARPTR^.NEXT := NIL ;
    UVARPTR^.KLASS := VARS ;
      UVARPTR^.VKIND := ACTUAL ;
      UVARPTR^.VADDR := 0 ;
      UVARPTR^.VLEV := 0 ;
    NEW(UFLDPTR) ;
    UFLDPTR^.NAME := '        ' ;
    UFLDPTR^.IDTYPE := NIL ;
    UFLDPTR^.NEXT := NIL ;
    UFLDPTR^.KLASS := FIELD ;
      UFLDPTR^.VADDR := 0 ;
    NEW(UPRCPTR) ;
    UPRCPTR^.NAME := '        ' ;
    UPRCPTR^.IDTYPE := NIL ;
    UPRCPTR^.FORWDECL := FALSE ;
    UPRCPTR^.NEXT := NIL ;
    UPRCPTR^.KLASS := PROC ;
      UPRCPTR^.PFDECKIND := DECLARED ;
        UPRCPTR^.PFLEV := 0 ;
        UPRCPTR^.PFNAME := 0 ;
        UPRCPTR^.EXTERNL := FALSE ;
    NEW(UFCTPTR) ;
    UFCTPTR^.NAME := '        ' ;
    UFCTPTR^.IDTYPE := NIL ;
    UFCTPTR^.NEXT := NIL ;
    UFCTPTR^.FORWDECL := FALSE ;
    UFCTPTR^.KLASS := FUNC ;
      UFCTPTR^.PFDECKIND := DECLARED ;
        UFCTPTR^.PFLEV := 0 ;
        UFCTPTR^.PFNAME := 0 ;
        UFCTPTR^.EXTERNL := FALSE ;
  END ;(* ENTERUNDECL *)

  BEGIN  (* INITIALIZE *)
    SETHSP ; (* OUTPUT TO HSP *)
    SETHSR  ; (* READ FROM HSR *)
    ERRFLAG   := FALSE ; (* INIT ERROR FLAG NO ERRORS YET *)
    LINECOUNT := 0 ;     (* ADD LINE NUMBERS *)
    FWPTR     := NIL   ;
    DP  := TRUE ;
    PRTERR    := TRUE ;
    ERRINX    := 0 ;
    LC        := LCAFTERMARKSTACK ;
    IC        := 0 ;
    ICN       := 0 ;
    CH        := ' ' ;
    ICH       := 32 ;

    MXINT10 := MAXINT DIV 10 ;
    NPROC := 1 ;
   (* INITIALIZE SETS *)
    CONSTBEGSYS    :=[ADDOP,INTCONST,STRINGCONST,IDENT] ;
    SIMPTYPEBEGSYS :=[LPARENT] + CONSTBEGSYS ;
    TYPEBEGSYS     :=[ARROW,PACKEDSY,ARRAYSY,RECORDSY,SETSY] +
                        SIMPTYPEBEGSYS ;
    TYPEDELS       :=[ARRAYSY,RECORDSY,SETSY] ;
    BLOCKBEGSYS    :=[CONSTSY,TYPESY,VARSY,PROCSY,FUNCSY,BEGINSY];
    SELECTSYS      :=[ARROW,PERIOD,LBRACK] ;
    FACBEGSYS      :=[INTCONST,STRINGCONST,IDENT,LPARENT,LBRACK,NOTSY] ;
    STATBEGSYS     :=[BEGINSY,IFSY,WHILESY,REPEATSY,CASESY,FORSY] ;
   (* INITIALIZE TABLES *)
    RW[ 0] := 'IF      ' ; RW[ 1] := 'DO      ' ;
    RW[ 2] := 'OF      ' ; RW[ 3] := 'IN      ' ;
    RW[ 4] := 'OR      ' ; RW[ 5] := 'TO      ' ;
    RW[ 6] := 'END     ' ; RW[ 7] := 'FOR     ' ;
    RW[ 8] := 'VAR     ' ; RW[ 9] := 'DIV     ' ;
    RW[10] := 'MOD     ' ; RW[11] := 'SET     ' ;
    RW[12] := 'AND     ' ; RW[13] := 'NOT     ' ;
    RW[14] := 'THEN    ' ; RW[15] := 'ELSE    ' ;
    RW[16] := 'TYPE    ' ; RW[17] := 'CASE    ' ;
    RW[18] := 'BEGIN   ' ; RW[19] := 'UNTIL   ' ;
    RW[20] := 'WHILE   ' ; RW[21] := 'ARRAY   ' ;
    RW[22] := 'CONST   ' ; RW[23] := 'REPEAT  ' ;
    RW[24] := 'RECORD  ' ; RW[25] := 'PACKED  ' ;
    RW[26] := 'EXTERN  ' ; RW[27] := 'DOWNTO  ' ;
    RW[28] := 'FORWARD ' ; RW[29] := 'PROGRAM ' ;
    RW[30] := 'FUNCTION' ; RW[31] := 'PROCEDUR' ;
    FRW[0] := 0   ;
    FRW[1] := 0   ;
    FRW[2] := 6   ;
    FRW[3] := 14  ;
    FRW[4] := 18  ;
    FRW[5] := 23  ;
    FRW[6] := 28  ;
    FRW[7] := 30  ;
    FRW[8] := 32  ; 
   (* INITIALIZE SYMBOLS *)
    RSY[ 0] := IFSY      ; RSY[ 1] := DOSY     ;
    RSY[ 2] := OFSY      ; RSY[ 3] := RELOP    ;
    RSY[ 4] := ADDOP     ; RSY[ 5] := TOSY     ;
    RSY[ 6] := ENDSY     ; RSY[ 7] := FORSY    ;
    RSY[ 8] := VARSY     ; RSY[ 9] := MULOP    ;
    RSY[10] := MULOP     ; RSY[11] := SETSY    ;
    RSY[12] := MULOP     ; RSY[13] := NOTSY    ;
    RSY[14] := THENSY    ; RSY[15] := ELSESY   ;
    RSY[16] := TYPESY    ; RSY[17] := CASESY   ;
    RSY[18] := BEGINSY   ; RSY[19] := UNTILSY  ;
    RSY[20] := WHILESY   ; RSY[21] := ARRAYSY  ;
    RSY[22] := CONSTSY   ; RSY[23] := REPEATSY ;
    RSY[24] := RECORDSY  ; RSY[25] := PACKEDSY ;
    RSY[26] := EXTERNSY  ; RSY[27] := DOWNTOSY ;
    RSY[28] := FORWARDSY ; RSY[29] := PROGSY   ;
    RSY[30] := FUNCSY    ; RSY[31] := PROCSY   ;
    FOR I :=0 TO 127 DO
      BEGIN
        SSY[I]  := OTHERSY ;
        SOP[I]  := NOOP
      END ;
    SSY[ORD('+')] := ADDOP     ;
    SSY[ORD('-')] := ADDOP     ;
    SSY[ORD('*')] := MULOP     ;
    SSY[ORD('/')] := MULOP     ;
    SSY[ORD('(')] := LPARENT   ;
    SSY[ORD(')')] := RPARENT   ;
    SSY[ORD('$')] := OTHERSY   ;
    SSY[ORD('=')] := RELOP     ;
    SSY[ORD(' ')] := OTHERSY   ;
    SSY[ORD(',')] := COMMA     ;
    SSY[ORD('.')] := PERIOD    ;
    SSY[ORD('"')] := OTHERSY   ;
    SSY[ORD('[')] := LBRACK    ;
    SSY[ORD(']')] := RBRACK    ;
    SSY[ORD(':')] := COLON     ;
    SSY[ORD('^')] := ARROW     ;
    SSY[ORD('<')] := RELOP     ;
    SSY[ORD('>')] := RELOP     ;
    SSY[ORD(';')] := SEMICOLON ;
   (* INITIALIZE OPERATORS *)
    FOR I := 0 TO 31 DO
      ROP[I] := NOOP ;
    ROP[ 3] := INOP  ;
    ROP[ 4] := OROP  ;
    ROP[ 9] := IDIV  ;
    ROP[10] := IMOD  ;
    ROP[12] := ANDOP ;
    SOP[ORD('+')] := PLUS  ;
    SOP[ORD('-')] := MINUS ;
    SOP[ORD('*')] := MUL   ;
    SOP[ORD('=')] := EQOP  ;
    SOP[ORD('/')] := IDIV  ;
    SOP[ORD('<')] := LTOP  ;
    SOP[ORD('>')] := GTOP  ;
   (* ENTER STANDARD NAMES AND STANDARD TYPES *)
    LEVEL      := 0   ;
    TOP        := 0   ;
    DISPLAY[0] := NIL ;
    ENTERSTDTYPES  ;
    STDNAMES ;
    ENTERSTNAMES ;
    ENTERUNDECL ;
    TOP := 1 ;
    LEVEL := 1 ;
    SAVETOP := 1 ;
    DISPLAY[1] := NIL ;
   (* COMPILE *)
    BEGINLINE ;
    INSYMBOL ;
    NEW(PROGPTR) ;
    PROGPTR^.IDTYPE := NIL ;
    PROGPTR^.NEXT := NIL ;
    PROGPTR^.KLASS := PROC ;
    PROGPTR^.PFDECKIND := DECLARED ;
    PROGPTR^.PFNAME := 0 ;
    PROGPTR^.NAME := '        ' ;
    IF SY = PROGSY
      THEN
        BEGIN
         (* PROGRAM NAME *)
          INSYMBOL ;
          IF SY <> IDENT
            THEN
              ERROR(2)
            ELSE
              PROGPTR^.NAME := ID ;
         (* SKIP INPUT/OUTPUT AFTER PROGRAM NAME *)
          WHILE SY <> SEMICOLON DO
            INSYMBOL ;
         (* SKIP SEMICOLON *)
          INSYMBOL ;
        END ;
   (* JUMP INTO BLOCK UNTIL END *)
    BLOCK(BLOCKBEGSYS + STATBEGSYS,PERIOD,PROGPTR) ;
    ENDLINE ; 
    WRITEOUT ;
	WRITELN('P9') ;
    IF ERRFLAG
      THEN
        BEGIN
	  TTYOUT ;
	  WRITELN('COMPILE ERRORS')
	END
  END.
